c 
c Runtime data partitioning 
c
c Initial data distribution - uniform blocks
c PARTI Data partitioning primitives are called to
c   a. partition arrays
c   b. partition loop iterations of sweep over edges
c   c. partition loop iterations of sweep over faces
c   d. remap arrays and loop iterations
c
      subroutine dopart( new_mynodes, tnode, 
     .      mynnodes, mynedges, mynfaces, nbnode,
     .      end_pts,bf_pts, volume,icsr_ptr, icsr_cols, 
     .      ripa, ref, len, filen)

      implicit none
#include"common.F"
c
      integer  new_mynodes(mnode)
      integer  mynnodes, mynedges, mynfaces, nbnode
      integer  end_pts(medge,2), bf_pts(mbface,4)
      real*8   volume(mnode)
      integer ripa(nge), ref(nge)
      integer icsr_ptr(nge), icsr_cols(nge)
      integer  len
      character *80 title,filen
c Locals
      integer mynodes(mnode), myedges(medge), myfaces(mbface)
      integer new_myedges(medge), new_myfaces(mbface)
c
      integer ncols
      integer nnode,nedge,nbface
      integer tnode,tedge,tbface,tbnode
      integer hashindex, init_rdg_hash_table
      integer  newnnode, newnedge, newnface
c
      integer nsched, esched, fsched, itable, ntable, rtable
      integer  sched, otable,  ref_per_iter

      logical build_translation_table

      integer i, mynode, n_dep, j

c
c--Read in Relevent local Data Structures in uniform blocks
c
      call readmesh(mnode,medge,mbface,
     .                  mynodes, myedges, myfaces,
     .                  mynnodes, mynedges, mynfaces,
     .                  tnode,tedge,tbface,tbnode,end_pts,
     .                  volume,bf_pts,
     .                  len, filen)
      call gsync()
      if (mynode() .eq. 0) print *, 'Finished Reading ...'
      call gsync()
c
c-- Setup a table for the initial array distribution
c
      itable = build_translation_table(1, mynodes, mynnodes)

      if (mynode() .eq. 0) print *, 'Partitioning nodes ...'
c
c  initialize hashtable for RDG generation
c
      hashindex = init_rdg_hash_table(mynedges)

c
c--The following loop(sweep over edges) has been selected for data partitioning.
c  Arrays flux, w, and cc are considered for graph generation and they are
c  to  be paritioned in a conforming manner.
c
c--Sweep over edges
c
c      do 500 i=1,noedge
c      n1       = end_pts(i,1)
c      n2       = end_pts(i,2)
c      a1       = (econs*w(n1,2) + econs*w(n1,3)
c     .                               + econs*w(n1,4)) / w(n1,1)
c      a2       = (econs*w(n2,2) + econs*w(n2,3)
c     .                               + econs*w(n2,4)) / w(n2,1)
c      qs       = (a1 + a2) / 2.0D0
cS1      flux(n1)    = flux(n1) + dabs(qs) + cc(n1)*econs
cS2      flux(n2)    = flux(n2) + dabs(qs) + cc(n2)*econs
c  500 continue
c
c--Here,statement S1 accesses array element n1 on the left hand side(lhs) and 
c  n1 and n2 on the right hand side(rhs). For graph generation,  we consider 
c  unique array accesses on the rhs and we ignore self-edges.
c  Hence, we pass only the list of n1 (end_pts(1,1)) as lhs access list and  
c  list of  n2 (end_pts(1,2) as rhs access list. Here, statement S2 will not
c  produce any new edges other than the edges produced by S1, so we dont
c  condiser S2 for local graph generation.  

      n_dep = 1
      call eliminate_dup_edges(hashindex, end_pts(1,1), 
     .       end_pts(1,2), n_dep, mynedges)
c
c-- merge local graphs to form a distributed graph 
c
      call generate_rdg(hashindex, mynodes, mynnodes, 
     .       icsr_ptr, icsr_cols, ncols)

      if (nge .lt. ncols) print *, 'Error: not enuf space', ncols
      if (nge .lt. icsr_ptr(mynnodes+1)) then
        print *, 'Error: not enuf space for csr_cols', icsr_cols(mynnodes+1)
      end if

c
c  call parallel partitioner
c
      call parallel_partitioner(mynodes, mynnodes, icsr_ptr,
     .           icsr_cols, rtable)

c
c-- remap array indices
c
      call remap(rtable, mynodes, nsched, new_mynodes, newnnode)

      if(mynode() .eq. 0) print *, ' '
      if (newnnode .gt. mnode) print *, 'Not enough space to hold nodes'

c
c-- build a translation table for the new array distribution
c
      ntable = build_translation_table(1, new_mynodes, newnnode)
c
c free tables and schedules
c
       call free_sched(nsched)
       call free_table(rtable)
c
c call loop iteration partitioner - 
c loop iterations are partitioned based on the new array distributions
c
      if(mynode()  .eq. 0) print *, 'Partitioning edge loop ...'

c-- generate list of right and left hand side indices accessed in the edge loop
      j = 1
      ref_per_iter = 2
      do i = 1, mynedges
       ref(j) = end_pts(i,1)
       j      = j + 1
       ref(j) = end_pts(i,2)
       j      = j + 1
      end do
c
c Partitiong edge loop iterations 
c
c-- generate runtime iteration processor assignment graph(ripa)
      call dref_rig(ntable, ref, mynedges, ref_per_iter,  ripa)
c-- partition loop iterations 
      call iteration_partitioner(ripa, mynedges, ref_per_iter, rtable)
c
      call remap(rtable, myedges,  esched, new_myedges, newnedge)
c
      if (newnedge .gt. medge) print *, 'Not enuf mem to hold edges', newnedge
c
c-- remap data associated with edges 
c
      call igather( end_pts, end_pts, esched)
      call igather( end_pts(1,2), end_pts(1,2), esched)
c

c free tables and schedules
c
       call free_sched(esched)
       call free_table(rtable)
c
c boundary face loop partitioning
c
      if(mynode()  .eq. 0) print *, 'Partitioning face loop ...'
c
c generate a list of unique indices accessed in each iteration
c
      ref_per_iter = 3
      j = 1
      do i = 1, mynfaces
       ref(j)   = bf_pts(i,1)
       ref(j+1) = bf_pts(i,2)
       ref(j+2) = bf_pts(i,3)
       j = j + 3
      end do

c generate ripa
      call dref_rig(ntable, ref, mynfaces, ref_per_iter,  ripa)
c partition loop iterations
      call  iteration_partitioner(ripa, mynfaces, ref_per_iter, rtable)

      call remap(rtable, myfaces,  fsched,  new_myfaces, newnface)
c
      if (newnface .gt. mbface) print *, 'Not enuf mem to hold faces', newnface
c
c-- remap data associated with boundary faces
c
      call igather( bf_pts, bf_pts, fsched)
      call igather(bf_pts(1,2), bf_pts(1,2), fsched)
      call igather(bf_pts(1,3), bf_pts(1,3), fsched)
      call igather(bf_pts(1,4), bf_pts(1,4), fsched)
c
c free tables and schedules
c
      call free_sched(fsched)
      call free_table(rtable)


      mynnodes = newnnode
      mynedges = newnedge
      mynfaces = newnface

      call free_table(ntable)

      end
