c-------------------------------------------------------------------------------
C
C Loops over edges and faces of an unstructured mesh. This is
C based on procedures from Dimitri Mavriplis' unstructured Euler
C solver.
C
c------------------------------------------------------------------------------

      subroutine loops(ledge,lnode,lbface,end_pts,nbf,
     .                w,p,flux,volume)
c
c--Compute Polyhedral Control Volumes
c
c-------------------------------
#include "common.F"
c-------------------------------
c
c
      common/node/ ntotnodes,nonode,noedge,nobnode,nobface
      common/sched/ lesched,lfsched,lfsched2
      common/offproc/ ne_off_proc,nf_off_proc,nf_off_proc2
c
      integer end_pts(ledge,2)
      integer nbf(lbface,4)
      integer iit(3)
      real*8  ssface
      real*8  flux(lnode),volume(lnode),fcons, econs, vols
      real*8  w(lnode,5),p(lnode)
      real*8  cc(mnode)
      real*8  cc1,cc2,cs1,cs2,a1,a2,qs
      real*8  sface,cs3,a3
      real*8  ftemp, flux_sum
      integer bshift, dialation, repetition
c
c--Initialize Time Step
c
      econs = 1.0D0
      fcons = 1.0D0
      ssface = 1.0D0

      do 50 i=1,nonode
      flux(i)   = 0.0D0
   50 continue
c
c Do all the Gathers  
c lesched --> schedule for loop over edges
c lfsched2 --> schedule for loop over boundary faces
c
      iit(1) = lesched
      iit(2) = lfsched2

      bshift = lnode
      dialation = 1
      repetition = 4
      call dmultigather(w(nonode+1,1),w(1,1),2,iit,bshift,dialation,repetition)

      repetition = 1
      call dmultigather(p(nonode+1),p(1),2,iit,bshift,dialation,repetition)
c
      do 160 i=1,nonode+ne_off_proc+nf_off_proc2
      cc(i) = p(i)/w(i,1)
  160 continue
c
c Zero-out buffer
c
      do 63 i = 1,ne_off_proc
        flux(nonode+i) = 0.0D0
   63 continue
c
c--Sweep over edges
c
      do 500 i=1,noedge
      n1       = end_pts(i,1)
      n2       = end_pts(i,2)
      a1       = (econs*w(n1,2) + econs*w(n1,3)
     .                               + econs*w(n1,4)) / w(n1,1)
      a2       = (econs*w(n2,2) + econs*w(n2,3)
     .                               + econs*w(n2,4)) / w(n2,1)
      qs       = (a1 + a2) / 2.0D0
      flux(n1)    = flux(n1) + dabs(qs) + cc(n1)*econs
      flux(n2)    = flux(n2) + dabs(qs) + cc(n2)*econs
  500 continue
c
      call dscatter_add(flux(nonode+1),flux(1),lesched)
c
      do 73 i = 1,ne_off_proc+nf_off_proc2
        flux(nonode+i) = 0.0D0
   73 continue
c
c--Sweep over boundary faces
c
      do 550 i=1,nobface
      n1       = nbf(i,1)
      n2       = nbf(i,2)
      n3       = nbf(i,3)
      cs1      = cc(n1) * ssface
      cs2      = cc(n2) * ssface
      cs3      = cc(n3) * ssface
      a1       = (fcons*w(n1,2) + fcons*w(n1,3)
     .                                + fcons*w(n1,4)) / w(n1,1)
      a2       = (fcons*w(n2,2) + fcons*w(n2,3)
     .                                + fcons*w(n2,4)) / w(n2,1)
      a3       = (fcons*w(n3,2) + fcons*w(n3,3)
     .                                + fcons*w(n3,4)) / w(n3,1)
c
      flux(n1)= flux(n1) + cs1 + dabs(a1) + (dabs(a1+a2) + dabs(a1+a3)) / 4.0D0
      flux(n2)= flux(n2) + cs2 + dabs(a2) + (dabs(a2+a1) + dabs(a2+a3)) / 4.0D0
      flux(n3)= flux(n3) + cs3 + dabs(a3) + (dabs(a3+a1) + dabs(a3+a2)) / 4.0D0
c
  550 continue
c
c Do Scatter
c
      call dscatter_addnc(flux(1),flux(1),lfsched)
c
c-- Checksum
c
      flux_sum = 0.0D0
      do 700 i=1,nonode
      flux(i)   = volume(i)/flux(i) * 3.0D0
      flux_sum = flux_sum + flux(i)
 700  continue
      call gdsum(flux_sum, 1, ftemp)
      if(mynode() .eq. 0) write(*,*), mynode(), flux_sum

      end
