#include <stdio.h>
#include <cube.h>
#include "parti.h"
#include "graph.h"


/***********************************************************************/
/* This file has functions to                                          */
/* initiallize hash table : rdg_hash_table_init()                      */
/* remove duplicaate edges: eliminate_dup_edges()                      */
/* merge local to form distributed RDG : generate_rdg()                */
/***********************************************************************/

/***********************************************************************/
/* Generates local CSR representation of dependency graph.             */
/*
on proc P:
Sequential
   do i=1,n
    n1 = nde(i,1)
    n2 = nde(i,2)
    y(n1) = x(n1) + x(n2) 
   end do

On processor:

  do i=1,n_local
   n1 = local_nde(i,1)
   n2 = local_nde(i,2)
    y(n1) = x(n1) + x(n2) +x(n3) 
   end do

Inputs:
f - left hand side indices
g - right hand side indicies
f = (nde(i,1) for all i assigned to P)
g = (nde(i,2),nde(i,3) for all i assigned to P)
n_dep = number of right hand side indices (2 - n2 and n3)
n_loop = number of local loop iterations
*/

/***********************************************************************/
/* eliminate_dup_edges marks edges between f(i) and g(i, j) where 
   0<j<n_dep[i], and 0<i<n_loop.
   It also removes duplicates edges, if any, using the table.

  -ravi 
*/
/***********************************************************************/

void
eliminate_dup_edges(htptr, f, g, n_dep,  n_loop)
int *f, *g, n_dep;
int n_loop;
HTABLE * htptr;
{
int i, j, k, l, m, loop;
int lhs, rhs;
HASHPTR   *htable, *hptr;
LISTPTR  *lptr, *cur_lptr;
int found; 
int rhs_count;
int temp, mynum;
FILE *fp;
char fname[20];
int n_rhs, n_lhs;


    n_rhs     = htptr->n_rhs;
    n_lhs     = htptr->n_lhs;
    htable    = htptr->hashtable;
    mynum     = mynode();
    rhs_count = 0;

    for(loop=0; loop<n_loop; loop++) {
      lhs = f[loop];
      for(k=0; k<n_dep; k++) {
          rhs = g[rhs_count++];  
          /* undirected graph  rhs -- lhs  */
          for(l=0; l<2; l++) {
             if (l==1) {
                temp = lhs;
                lhs  = rhs;
                rhs  = temp;
             }
               
             if (lhs != rhs) {
                hptr = &htable[lhs%SIZE];
                found = FALSE;
                while (hptr->next != NULL ) {
                  hptr = hptr->next;
                  if (hptr->index == lhs) {
                    lptr = hptr->next_list;
                    while (lptr != NULL){
                      if (lptr->index == rhs) {
                        found = TRUE;
                        break;
                      }
                      cur_lptr = lptr;
                      lptr     = lptr->next;
                    }
                    if (!found) {
                      cur_lptr->next  = (LISTPTR *)
                                         my_malloc(sizeof(struct list_t));
                      cur_lptr        = cur_lptr->next;
                      cur_lptr->index = rhs; 
                      cur_lptr->next  = NULL; 
                      found           = TRUE;
                      n_rhs++;
                      break;
                    }
                  }
                }
                if (!found) { /* this lhs has not been encountered yet */
                  hptr->next  = (HASHPTR *) 
                                  my_malloc(sizeof(struct hash_struct));    
                 hptr        = hptr->next;
                 hptr->index = lhs;
                 hptr->next  = NULL;
                 cur_lptr    = hptr->next_list 
			     = (LISTPTR *) my_malloc(sizeof(struct list_t));
                 cur_lptr->index = rhs;
                 cur_lptr->next  = NULL;
                 n_lhs++;
                 n_rhs++;
                }
             }
           }
         }
    }

  /* update the values */
  htptr->n_rhs  = n_rhs;
  htptr->n_lhs  = n_lhs;
}



/**  fortran  callable function */
void eliminate_dup_edges_(hashindex, f, g, n_dep, n_loop)
int *f, *g, *n_dep, *n_loop;
int *hashindex;
{
int i, j,  
    n_dep_val, 
    n_loop_val;
FILE *fp;
int count;
HTABLE *hashtab;

    n_dep_val  = *n_dep;
    n_loop_val = *n_loop;
    count = 0;
    for (j=0; j<n_loop_val; j++)  {
      f[j]--;
      for (i=0; i<n_dep_val; i++) {
	    g[count++]--; 
      }
    }
    hashtab = (HTABLE *) get_hashtable_ptr(*hashindex);

    eliminate_dup_edges(hashtab, f, g, n_dep_val, n_loop_val);

    count = 0;
    for (j=0; j<n_loop_val; j++)  {
      f[j]++;
      for (i=0; i<n_dep_val; i++) {
	    g[count++]++; 
      }
    }
}



/*****************************************************************************/
/* generate_rdg flattens the hash table to lists ( local RDG in lists).      */
/* Then it calls scatter_append to merge the local RDG to form distributed   */
/* RDG. The  distributed graph is passed to the partitioner.                 */
/*****************************************************************************/

/*****************************************************************************
* row_ind - initial global indcies assigned
* nrows    - size of row_ind
* OUTPUT:
* RDG -
* csr_cols  - list of indices associated with row_ind
* csr_ptr  - pointer into csr_ind
* ncols     - size of csr_cols
*****************************************************************************/


void generate_rdg(htptr, row_ind, nrows, csr_ptr, csr_cols, ncols) 
HTABLE *htptr;
int *row_ind, nrows, *csr_ptr, *csr_cols, *ncols;
{
int i, j, n;
int new_ind, cur_ref;
HASHPTR *hptr;
HASHPTR *htable;
LISTPTR *lptr;
TTABLE *ttable;
SCHED *sched;
int *lhs_ind, *rhs_ind, *rhs_ptr;
int *local, *proc;
int nrow_ind ;
int *xadj, *adj;
int pfac, table_size;
int n_rhs, n_lhs;

char fname[20];
FILE *fp;
int count, nproc;
int *my_count, *ptr;
double dclock(), st_clock, end_clock;


/****************************************************************************/
/* lhs_ind - list of global RDG graph indices known by processor            */
/* rhs_ind - list of global RDG graph indices known by processor - associated */
/* with lhs_ind (in CSR format)                                               */
/* rhs_ptr - ptr into CSR formatted my_rhs_ind array                          */
/******************************************************************************/

  n_rhs   = htptr->n_rhs;
  n_lhs   = htptr->n_lhs;
  htable  = htptr->hashtable;
  lhs_ind = (int *) malloc(sizeof(int)*n_lhs);
  rhs_ind = (int *) malloc(sizeof(int)*n_rhs);
  rhs_ptr = (int *) malloc(sizeof(int)*(n_lhs+1));

  nrow_ind = nrows;
  new_ind = cur_ref = 0;

  for (i=0; i<SIZE; i++) {
     hptr = &htable[i];
     while(hptr->next != NULL) {
       hptr = hptr->next;
       lhs_ind[new_ind] = hptr->index;
       rhs_ptr[new_ind++]   = cur_ref;
       lptr = hptr->next_list;
       while(lptr != NULL) {
          rhs_ind[cur_ref++] = lptr->index;
          lptr = lptr->next;
       }
     }
  }
  rhs_ptr[new_ind] = cur_ref;

   my_free();
   free(htable);
   free(htptr);


   
   /* pre-processing for scatter_append */

   local = (int *) malloc(sizeof(int)*n_lhs);
   proc  = (int *) malloc(sizeof(int)*n_lhs);

   /*
    since block distribution is used  no need for ttable and dereference 
   */
   
   /*
   table_size = row_ind[nrow_ind-1];
   gihigh(&table_size, 1, &i);
   nproc = numnodes();
   table_size = (table_size) / nproc;
   for(i=0; i<n_lhs; i++) {
     proc[i]  = lhs_ind[i]/table_size;
     if (proc[i] == nproc) proc[i]--;
     local[i] = lhs_ind[i]- (table_size*proc[i]);
   }
   */
   
   
   ttable = (TTABLE *) build_translation_table(BLOCKED, row_ind, nrow_ind);
   dereference(ttable, lhs_ind, local, proc, n_lhs);
  
   free_table(ttable);
   
   sched  = (SCHED *) schedule (local, proc, n_lhs, 1);


   scatter_append(sched, rhs_ptr, rhs_ind, nrow_ind, csr_ptr, csr_cols, ncols);
   
   free(lhs_ind); free(rhs_ind); free(rhs_ptr); free(local); free(proc);
   free_sched(sched);

}


/* fortran callable */
void generate_rdg_(hashindex, row_ind, nrow_ind, csr_ptr, csr_cols, ncols)
int *hashindex, *row_ind, *nrow_ind, *csr_ptr, *csr_cols;
int *ncols;
{
   int i;
   HASHPTR *htable;

   for(i=0; i<*nrow_ind; i++) {
       row_ind[i]--;
   }
  htable = (HASHPTR *) get_hashtable_ptr(*hashindex);
  generate_rdg( htable, row_ind, *nrow_ind, csr_ptr, csr_cols, ncols);

   for(i=0; i<*nrow_ind; i++) {
       row_ind[i]++;
   }
}



/******************************************************************/
/* Hash table initialization                                      */
/* init_rdg_hash_table- allocates memory for hash table based on */
/* size and returns a pointer to it.                              */
/******************************************************************/

HTABLE * init_rdg_hash_table(size)
int size;
{
    int i;
    HASHPTR *htable;
    HTABLE *htptr;

    my_alloc_init(size);
    htptr            = (HTABLE *)malloc(sizeof(struct hash_table));
    htptr->hashtable = (HASHPTR *)malloc(sizeof(struct hash_struct)*SIZE);
    htable           = htptr->hashtable;
    htptr->n_rhs     =  0;
    htptr->n_lhs     =  0;
    for(i=0;i<SIZE;i++){
      htable[i].next     = NULL;
      htable[i].next_list = NULL;
    }

    return htptr;
}

/* fortran callable */
int init_rdg_hash_table_(size)
int *size;
{
     return (int ) register_hashtable(init_rdg_hash_table(*size));
}
