/*
* data remapping primitive
*/
#include <stdio.h>
#include <cube.h>
#include "../include/parti.h"

#define MAP_MSG  10000

/******************************************************************/
/* 							          */
/* given a distributed processor assignment array(color),         */
/* redistribute the initial mapping (cur_local_ind)               */
/* and return a  schedule (sched) to remap any array that might   */
/* be aligned to this particular initial mapping(cur_local_ind)   */
/*                                                                */
/******************************************************************/

remap_old(color, cur_local_ind, cur_n_data,  sched, new_local_ind, newndata)
int *color, *cur_local_ind;
int cur_n_data;
SCHED **sched;
int *new_local_ind, *newndata;
{ 
 int i, j, count;
 int mynum, nproc;
 int *pcount, *bcount, psize;
 int *buffer, **buckets;
 int *wbuf;
 int recv_n;
 int pid, proc, bsize, isize;
 int *Local, *Proc;
 TTABLE *otable;
 int intcompare();
 int remap_n_data;

 
 char fname[10];
 FILE *fp;

 mynum = mynode();
 nproc = numnodes();
 pid   = mypid();

 otable  = build_translation_table(BLOCKED, cur_local_ind, cur_n_data);

 buckets = (int **) malloc(sizeof(int *)*nproc);
 pcount  = (int *) malloc(sizeof(int *)*nproc);
 bcount  = (int *) malloc(sizeof(int *)*nproc);
 wbuf    = (int *) malloc(sizeof(int *)*nproc);
 buffer  = (int *) malloc(sizeof(int)*(cur_n_data));
 
 for(i=0; i<nproc; i++) pcount[i] = 0;

 for(i=0; i<cur_n_data; i++) { 
     pcount[color[i]]++;
 }

 psize = 0;
 for(i=0; i<nproc; i++) {
   buckets[i] = &buffer[psize];
   psize    += pcount[i];
 }

 for(i=0; i<nproc; i++) bcount[i] = 0;

 for(i=0; i<cur_n_data; i++) {
   proc = color[i];
   buckets[proc][bcount[proc]++] = cur_local_ind[i];
 }

 gisum(pcount, nproc, wbuf);

 recv_n = pcount[mynum];

 for(i=0; i<nproc; i++) {
   if (pcount[i] > 0) {
     csend(MAP_MSG, buckets[i], sizeof(int)*bcount[i], i, pid); 
   }
 } 

 remap_n_data = recv_n;

 psize = 0;
 while (recv_n > 0) {
   cprobe(MAP_MSG);
   bsize = infocount();
   isize = bsize/sizeof(int);
   crecv(MAP_MSG, &new_local_ind[psize], bsize);
   psize  += isize;
   recv_n -= isize;
 }
 
 free(buckets); free(pcount); free(buffer); free(wbuf); free(bcount);

 Local  = (int *) malloc(sizeof(int)*(remap_n_data));
 Proc   = (int *) malloc(sizeof(int)*(remap_n_data));

 qsort(new_local_ind, remap_n_data, sizeof(int), intcompare);

 dereference(otable, new_local_ind, Local, Proc, remap_n_data);
 *sched = schedule(Local, Proc, remap_n_data, 1);
 *newndata = remap_n_data;

 free(Local); free(Proc); 
 free_table(otable);

}

static int
intcompare(x, y)
int *x, *y;
{
 return (*x-*y);
}


/******************************************************************/
/* 							                                      */
/* given a distributed processor assignment distributed table(table),*/
/* redistribute the initial mapping (cur_local_ind)               */
/* and return a  schedule (sched) to remap any array that might   */
/* be aligned to this particular initial mapping(cur_local_ind)   */
/*                                                                */
/******************************************************************/

remap(table, cur_local_ind,  sched, new_local_ind, newndata)
TTABLE *table;
int  *cur_local_ind;
SCHED **sched;
int *new_local_ind, *newndata;
{ 
 int i, j, count;
 int mynum, nproc;
 int *pcount, *bcount, psize;
 int *buffer, **buckets;
 int *wbuf;
 int recv_n;
 int pid, proc, bsize, isize;
 int *Local, *Proc;
 int *color;
 TTABLE *otable;
 int intcompare();
 int remap_n_data;
 int cur_n_data;

 
 char fname[10];
 FILE *fp;

 mynum = mynode();
 nproc = numnodes();
 pid   = mypid();
 color = table->proc;
 cur_n_data = table->table_size;

 otable  = build_translation_table(BLOCKED, cur_local_ind, cur_n_data);

 buckets = (int **) malloc(sizeof(int *)*nproc);
 pcount  = (int *) malloc(sizeof(int *)*nproc);
 bcount  = (int *) malloc(sizeof(int *)*nproc);
 wbuf    = (int *) malloc(sizeof(int *)*nproc);
 buffer  = (int *) malloc(sizeof(int)*(cur_n_data));
 
 for(i=0; i<nproc; i++) pcount[i] = 0;

 for(i=0; i<cur_n_data; i++) { 
     pcount[color[i]]++;
 }

 psize = 0;
 for(i=0; i<nproc; i++) {
   buckets[i] = &buffer[psize];
   psize    += pcount[i];
 }

 for(i=0; i<nproc; i++) bcount[i] = 0;

 for(i=0; i<cur_n_data; i++) {
   proc = color[i];
   buckets[proc][bcount[proc]++] = cur_local_ind[i];
 }

 gisum(pcount, nproc, wbuf);

 recv_n = pcount[mynum];

 for(i=0; i<nproc; i++) {
   if (pcount[i] > 0) {
     csend(MAP_MSG, buckets[i], sizeof(int)*bcount[i], i, pid); 
   }
 } 

 remap_n_data = recv_n;

 psize = 0;
 while (recv_n > 0) {
   cprobe(MAP_MSG);
   bsize = infocount();
   isize = bsize/sizeof(int);
   crecv(MAP_MSG, &new_local_ind[psize], bsize);
   psize  += isize;
   recv_n -= isize;
 }
 
 free(buckets); free(pcount); free(buffer); free(wbuf); free(bcount);

 Local  = (int *) malloc(sizeof(int)*(remap_n_data));
 Proc   = (int *) malloc(sizeof(int)*(remap_n_data));

 qsort(new_local_ind, remap_n_data, sizeof(int), intcompare);

 dereference(otable, new_local_ind, Local, Proc, remap_n_data);
 *sched = schedule(Local, Proc, remap_n_data, 1);
 *newndata = remap_n_data;

 free(Local); free(Proc); 
 free_table(otable);

}
