#include <cube.h>
#include <stdio.h>
#include "parti.h"

#define APP_LIST      100
#define APP_LIST_SIZE 300

/**************************************************************/
/* scatter append - takes in local graph in CSR format        */
/* and returns distributed RDG in CSR format                  */
/*                                                            */
/* Assume that, of an array of size 4, the indices 1, and 2,  */
/* are assigned processor 1 & indices 3and 5   are assigned   */
/* to processor 2.                                            */
/*  suppose the following are  the f and g                    */
/*      proc1		         proc2                        */
/*    f   g                    f   g                          */
/*    1 - 2, 3                 3 -  4, 1                      */
/*    2 - 3, 4                 4 -  1, 2                      */
/*   after generate_rdg (local RDG)                           */
/*    1 -> 2, 3                3 ->  4, 1                     */
/*    2 -> 1, 3, 4             4 ->  3, 1, 2                  */
/*    3 -> 1, 2                1 ->  3, 4                     */
/*    4 -> 2                   2 ->  4                        */
/* after scatter append: (distributed  RDG)                   */ 
/*    1  -> 2, 3, 4             3 ->  1, 2, 4                 */
/*    2  -> 1, 3, 4             4 ->  1, 2, 3                 */
/**************************************************************/

/* 05-26-91 */
/***************************************************/
/* IN:                                             */
/* sched: schedule (regular)                       */
/* ia:  row pointer in CSR                         */
/* ja: global column indices in CSR                */
/* OUT:                                            */
/* hum_ia: row pointer in CSR                      */
/* hum_ja: global  column indices in CSR           */
/***************************************************/
void scatter_append(sched, ia, ja, nrow, hum_ia, hum_ja, nhum_ia)
SCHED *sched;
int  *ia, *ja;
int nrow;
int  *hum_ia, *hum_ja, *nhum_ia;
{
    int count, tcount, np, i, j, k, m;
    int *l;
    int **slist, *buffer;
    int *sizbuf;
    int **lists;
    int *store, *recvbuf;
    int *sbuf;
    int *curptr;
    int sbuf_size, ndata;
    int node, *listsize;
    int lsize, procto, mynumber;
    struct com_pattern *pattern;
    int *temp;
    int cur_perm;

    int current;
    int first;
    int intcompare();

    int pid, *mybuf;

    np      = numnodes();
    mynumber= mynode();
    ndata   = sched->ndata;
    pattern = sched->sr;
    slist   = (int **) malloc(sizeof(int *)*np);
    buffer  = (int *) malloc(sizeof(int)*ndata);
    sizbuf  = (int *) malloc(sizeof(int)*nrow);

    /*** findout the size of list of lists to be recvd  ***/
    count = 0;
    for(i=0; i<np; i++) {
      slist[i] = &buffer[count]; 
      count   += pattern->fetch_n[i];
    }
 
    sbuf_size = 0;
    for(i=0; i<sched->ndata; i++) {
      buffer[i] = ia[sched->perm[i]+1]-ia[sched->perm[i]];
      sbuf_size += buffer[i];
    }
    for(i=0; i<nrow; i++) {
      sizbuf[i] = 0;
    }
    
    /* add list sizes */
    for(i=0;i<np;i++){
      if(pattern->fetch_n[i]>0){
        csend(1000,slist[i],
              sizeof(int)*pattern->fetch_n[i],i,pid);
      }
    }

    for(i=0;i<pattern->n_msgs_to_send;i++){
      cprobe(1000);
      node = infonode();
      count = infocount();
      mybuf = (int *) malloc(count);
      crecv(1000,mybuf,
          pattern->buffer_size*sizeof(int));
      l = pattern->send_list[node];
      for(j=0;j<pattern->send_n[node];j++){
        sizbuf[l[j]] += mybuf[j];
      }
      free(mybuf);
     }

    
    /* total size of lists to be recd */
    count = 0;
    for(i=0; i<nrow; i++) {
      count += sizbuf[i];
    }


    /* allocate memory for the list to be recd incl locals
        and initialize curptr(list size) */
    recvbuf = (int *) malloc(count*sizeof(int));
    store   = (int *) malloc(count*sizeof(int));
    curptr  = (int *) malloc((nrow+1)*sizeof(int));
    lists   = (int **) malloc(nrow*sizeof(int));
     

    for(count=0,i=0; i<nrow; i++){
      lists[i]  = &recvbuf[count];
      count    += sizbuf[i];
    }
		
    for(i=0; i<nrow; i++){
      curptr[i] = 0;
    }

    sbuf     = (int *) malloc(sizeof(int)*sbuf_size);
    listsize = (int *) malloc(sizeof(int)*pattern->buffer_size);
    cur_perm = 0;
    for (i=0; i<np; i++) {
      if (pattern->fetch_n[i] > 0) {
         /* send list size */
         csend(APP_LIST_SIZE+mynumber, slist[i], 
              sizeof(int)*pattern->fetch_n[i], i, 0); 

         /* pack the list of arrays  to be sent */
	 
         l = pattern->fetch_list[i];
         for(count=0,k=0; k<pattern->fetch_n[i]; k++) {


         for(j=ia[sched->perm[cur_perm]]; j<ia[sched->perm[cur_perm]+1]; j++) {
             sbuf[count] = ja[j];
             count++;
           }
           cur_perm++;
         }
	
         csend(APP_LIST+mynumber, sbuf, count*sizeof(int), i, 0); 
      }
    }
    
    tcount = 0;
    for (i=0; i<np; i++) {
      if (pattern->send_n[i] > 0) {
         /* recv the individual list sizes */
        crecv(APP_LIST_SIZE+i, listsize, 
                 sizeof(int)*pattern->send_n[i], 0); 
        lsize = 0;
        for(j=0; j<pattern->send_n[i]; j++) 
             lsize += listsize[j];
        /* recv the actual list  */
	
        crecv(APP_LIST+i, store, lsize*sizeof(int), 0); 
	
        /* append without duplicates */
        l = pattern->send_list[i];
        for (count=0,j=0; j<pattern->send_n[i]; j++) {
	    first = l[j];
          for (m=0; m<listsize[j]; m++) {
              lists[first][curptr[first]] = store[count];
              curptr[first] ++;
              count++;
	      tcount++;
          }
       }
     }
   } 

   hum_ia[0] = 0;
   count = 0;
   for(i=0; i<nrow; i++) {
     qsort(lists[i], curptr[i], sizeof(int), intcompare);
     current = lists[i][0];
     for(j=1; j<curptr[i]; j++) {
	if (current  != lists[i][j])  {
	   hum_ja[count] = current;
	   current = lists[i][j];
	   count++;
	}
     }
     hum_ja[count] = current;
     count++;
     hum_ia[i+1]   = count;
   }
   *nhum_ia = count;

   free(store); free(buffer); free(sizbuf);
   free(sbuf); free(curptr);
   free(slist); free(listsize);
   free(recvbuf); free(lists);
}



static    int intcompare(i,j)
	  int *i, *j;
{
return(*i - *j);
}
