###################################################################
#                                                                 #
# PARTI : Parallel Automated Runtime Tools from ICASE             #
# This is the top level makefile for PARTI. It recursively        #
# calls makefiles in the subdirectories.                          #
#                                                                 #
###################################################################
# 
#  NFLAG = to match type of cube user has.
#          -node        (for vanilla iPSC/2)
#          -node -sx    (for iPSC/2 with wtk)
#          -node        (for iPSC/860)
#  NARC = archiver (ar if iPSC/2, ar860 if iPSC/860)
#  LIB = where library should be intalled
#  INCL = where include files should be installed
#
###################################################################
#
# Top-level Makefile to produce PARTI and COUPLER archives for
# various types of target machines such as iPSC/860, CM5 and PVM.
#	Bongki Moon, Jan/25/93
#

ifeq ($(TARGET), SYRIPSC)
        CC     = icc
        F77    = if77
        AR     = ar860
        CFLAGS = -node -O2
endif
ifeq ($(TARGET), INTEL)
        CC     = icc
        F77    = if77
        AR     = ar860
        CFLAGS = -node -O2
endif
ifeq ($(TARGET), SYRCM5)
        CC     = cc
        F77    = /usr/lang/f77
        CFLAGS = -O3 -dalign
endif
ifeq ($(TARGET), CM5)
        CC     = cc
        F77    = /usr/lang/f77
        CFLAGS = -O3
endif
ifeq ($(TARGET), PVM)
        CC     = cc
        F77    = /usr/lang/f77
        AR     = ar
        CFLAGS = -O
endif
ifeq ($(TARGET), EXPRESS)
        CC     = xcc
        F77    = xf77
        AR     = ar
        CFLAGS = -kcubix
endif

#LIB    = ./lib
#INCL   = ./include 
PARTIHOME = $(PWD)
PARTI  = parti.$(TARGET).a
COUPLER= coupler.$(TARGET).a
PARTITIONER= partitioner.$(TARGET).a


all: target_name parti
.PHONY: all target_name clean

parti:
	@echo "Making $(PARTI)"
	cd src; $(MAKE)
#	@echo "Making $(COUPLER)"
#	cd msrc; $(MAKE)
#	@echo "Making $(PARTITIONER) -- EIGEN PARTITIONER"
#	cd partitioner; $(MAKE) 

target_name:
	@if [ x$(TARGET) != xINTEL -a x$(TARGET) != xSYRIPSC -a \
	      x$(TARGET) != xCM5 -a x$(TARGET) != xSYRCM5 -a \
	      x$(TARGET) != xPVM -a x$(TARGET) != xEXPRESS ]; then \
		echo 'The TARGET variable is not properly specified.'; \
		stop; \
	fi

clean: target_name
	cd src; $(MAKE) clean
#	cd msrc; $(MAKE) clean
#	cd partitioner; $(MAKE) clean








