/*-------------------------------------------------------------------
 * This file is written by Tseng-Hui Lin. If you have any question,
 * please send e-mail to:
 *     thlin@top.cis.syr.edu  or  thlin@nova.npac.syr.edu
 *
 * Last modify : 4/14/1993
 *-------------------------------------------------------------------
 */

/*
 * This file gives the name change for most communication system calls
 * This makes other files do not need to change when porting this 
 * system to other machines.
 */

#include <cm/cmmd.h>
#include <cm/cmmd-io.h>
#include "port.h"

#define SYNC_TYPE_BASE (0X40000000)
#define TIMER (63)
/* Time of CMMD_node_timer_stop() + CMMD_node_timer_busy() + CMMD_node_timer_start() */
#define TIME_GAP (0.0000275)
static double time_gap;

/* General communication calls for all systems */
void SYR_port_init()
{
    CMMD_fset_io_mode(stdin, CMMD_sync_bc);
    CMMD_fset_io_mode(stdout, CMMD_independent);
    CMMD_fset_io_mode(stderr, CMMD_independent);
    CMMD_node_timer_stop(TIMER);
    CMMD_node_timer_clear(TIMER);
    CMMD_node_timer_start(TIMER);
    time_gap = 0;
}

int SYR_mynode()
{
    return(CMMD_self_address());
}

int SYR_numnodes()
{
    return(CMMD_partition_size());
}

void SYR_blk_send(type, buf, len, dest)
int type, len, dest;
char *buf;
{
    CMMD_send_block(dest, type, buf, len);
}

void SYR_nonblk_send(type, buf, len, dest)
int type, len, dest;
char *buf;
{
    CMMD_send_noblock(dest, type, buf, len);
}

void SYR_sync_send(type, buf, len, dest)
int type, len, dest;
char *buf;
{
    CMMD_send_block(dest, type, buf, len);
}

int SYR_async_send(type, buf, len, dest)
int type, len, dest;
char *buf;
{
    return(CMMD_send_async(dest, type, buf, len, NULL));
}

void SYR_recv(type, buf, len, src)
int type, len, src;
char *buf;
{
    CMMD_receive(src, type, buf, len);
}

void SYR_sync_recv(type, buf, len, src)
int type, len, src;
char *buf;
{
    CMMD_receive(src, type, buf, len);
}

int SYR_async_recv(type, buf, len, src)
int type, len, src;
char *buf;
{
    return(CMMD_receive_async(src, type, buf, len, NULL));
}

void SYR_send_recv(stype, sbuf, slen, dest, rtype, rbuf, rlen, src)
int stype, slen, dest, rtype, rlen, src;
char *sbuf, *rbuf;
{
    CMMD_send_and_receive(src, rtype, rbuf, rlen, dest, stype, sbuf, slen);
}

void SYR_sync_send_recv(stype, sbuf, slen, dest, rtype, rbuf, rlen, src)
int stype, slen, dest, rtype, rlen, src;
char *sbuf, *rbuf;
{
    CMMD_send_and_receive(src, rtype, rbuf, rlen, dest, stype, sbuf, slen);
}

void SYR_swap(stype, sbuf, slen, dest, rtype, rbuf, rlen)
int stype, slen, dest, rtype, rlen;
char *sbuf, *rbuf;
{
    CMMD_send_and_receive(dest, rtype, rbuf, rlen, dest, stype, sbuf, slen);
}

int SYR_recv_len()
{
    return(CMMD_bytes_received());
}

int SYR_recv_sender()
{
    return(CMMD_msg_sender());
}

int SYR_recv_type()
{
    return(CMMD_msg_tag());
}

void SYR_async_msg_wait(msgid)
int msgid;
{
    CMMD_msg_wait(msgid);
    CMMD_free_mcb(msgid);
}

int SYR_async_msg_done(msgid)
int msgid;
{
    return(CMMD_msg_done(msgid));
}

void SYR_concat(elem, elem_len, whole)
char *elem, *whole;
int elem_len;
{
    int work;
    CMMD_concat_with_nodes(elem, whole, elem_len);
}

/* Global synchronous operation */

void SYR_gsync()
{
    CMMD_sync_with_nodes();
}

/* Time routines */

double SYR_clock()
{
    double time_temp;

    CMMD_node_timer_stop(TIMER);
    time_temp = CMMD_node_timer_busy(TIMER);
    CMMD_node_timer_start(TIMER);
    time_temp -= time_gap;
    time_gap += TIME_GAP;
    return(time_temp);
}

/*----------------------------------------------------------------*/
/* Specific communication calls for CM5 */


/*
 * For compatiblity of higher level PARTI use
 */

int gihigh(x, count, work)
int *x, count, *work;
{
    int i;

    for (i = 0; i < count; i++) {
        x[i] = CMMD_reduce_int(x[i], CMMD_combiner_max);
    }
}

double gdsum(x, count, work)
double *x, *work;
int count;
{
    int i;

    for (i = 0; i < count; i++) {
        x[i] = CMMD_reduce_double(x[i], CMMD_combiner_dadd);
    }
}

void gsync()
{
    CMMD_sync_with_nodes();
}

int numnodes()
{
    return(CMMD_partition_size());
}

int mynode()
{
    return(CMMD_self_address());
}
