/*-------------------------------------------------------------------
 * This file is written by Tseng-Hui Lin. If you have any question,
 * please send e-mail to:
 *     thlin@top.cis.syr.edu  or  thlin@nova.npac.syr.edu
 *
 * Last modify : 9/26/1993
 *-------------------------------------------------------------------
 */

#include <stdio.h>
#include <math.h>
#include "comm_int.h"

/*
 * Global (communication part) data secssion
 */

static SYR_csched_type SYR_curr_csched_type;
int SYR_debug = 0;
int SYR_nproc, SYR_myid;
boolean SYR_mtype_odd = FALSE;
static int SYR_mtype_msg;
static boolean need_init = TRUE;

/*
 * Global (this file) data secssion
 */

static SYR_com_pattern *SYR_comm_schedule(/* data_len, csched_type */);
static void SYR_cschedule_AC(/* len_tab, csched */);
static void SYR_cschedule_LP(/* len_tab, csched */);
static void SYR_cschedule_RS_N(/* len_tab, csched */);
static void SYR_mtype_flip(/*  */);

/*
 * NICE calling interfaces.
 */

void NICE_init()
{
    SYR_comm_init();
}

void nice_init_()
{
    NICE_init();
}

void NICE_set_sched_type(type)
int type;
{
    SYR_set_csched_type(type);
}

void nice_set_sched_type_(type)
int *type;
{
    NICE_set_sched_type(*type);
}

/*
 * This function set current communication scheduling
 * type and return the previous type.
 */

SYR_csched_type SYR_set_csched_type(type)
SYR_csched_type type;
{
    SYR_csched_type old_type;

    old_type = SYR_curr_csched_type;
    SYR_curr_csched_type = type;
    return(old_type);
}

/* 
 * Interface with upper level scheduling program.
 * This function create an internal data structure for communication and
 * return the pointer of this structure for latter use.
 */

SYR_com_pattern *schedule_gather(n, local_list, maxdim)
int *n, **local_list;
int maxdim;
{
    int i;
    int sl_size, *sl_p, *itempp;
    SYR_com_pattern *com_ptn;

    SYR_dbg_enter("schedule_gather");
    if ((com_ptn = SYR_comm_schedule(n, SYR_curr_csched_type)) != NULL) {
        com_ptn -> fetch_list = local_list;
        /* variable n is used in old version of schedule_gather to be */
        /* part of com_ptn. Since we don't use it as part of com_ptn  */
        /* we free it here to match the function of old version.      */
        free(n);

        /* Creat send_list */
        aloc_ary_a(sl_p, int, com_ptn -> tot_send_n);
        aloc_ary(com_ptn -> send_list, int *, SYR_nproc);
        for (i = 0, sl_size = 0; i < SYR_nproc; i++) {
            com_ptn -> send_list[i] = &sl_p[sl_size];
            sl_size += com_ptn -> send_n[i];
        }
        SYR_comm(local_list, com_ptn -> send_list, com_ptn -> s_mark,
            com_ptn -> s_num_stages, com_ptn -> csched_type, sizeof(int));
        for (i = 0; i < com_ptn -> fetch_n[SYR_myid]; i++) {
            com_ptn -> send_list[SYR_myid][i] = com_ptn -> fetch_list[SYR_myid][i];
        }

    }
    return(com_ptn);
}

/*
 * This is initialization routine of communication part.
 */

void SYR_comm_init()
{
    SYR_dbg_enter("SYR_comm_init");
    SYR_port_init();
    SYR_nproc = SYR_numnodes();
    SYR_myid = SYR_mynode();
    SYR_mtype_odd = FALSE;
    SYR_machine_init();
    SYR_curr_csched_type = SYR_SCH_AC;
    SYR_dbg_exit();
}

#define transpose(imat, dim) \
{ \
    int _i, _j, _itemp; \
 \
    for (_i = 0; _i < dim; _i++) { \
        for (_j = _i + 1; _j < dim; _j++) { \
            _itemp = imat[_i][_j]; \
            imat[_i][_j] = imat[_j][_i]; \
            imat[_j][_i] = _itemp; \
        } \
    } \
} \

/*
 * This function organizes the communication pattern, schedule
 * the order of sending/receiving data buffer and recores the
 * information in the schedule information table.
 */

static SYR_com_pattern *SYR_comm_schedule(data_len, csched_type)
int *data_len;
SYR_csched_type csched_type;
{
    boolean csched_err;
    SYR_com_pattern *csched;
    int self_len, **len_table;
    int i, j;
    int itemp;

    SYR_dbg_enter("SYR_comm_schedule");
    csched_err = FALSE;

    /* Schedule communication only. Do not scheule self send/receive */
    self_len = data_len[SYR_myid];
    data_len[SYR_myid] = 0;
    /* creat communication schedule table */
    aloc_ary(csched, SYR_com_pattern, 1);
    /* Allocate communication schedule table */
    aloc_mat(len_table, int, SYR_nproc, SYR_nproc);
    /* Get communication length from all other nodes */
    /* Concatnate with others to get whole communication patterns */
    SYR_concat(data_len, sizeof(int) * SYR_nproc, len_table[0]);

    /* Do communication scheduling */
    csched -> csched_type = csched_type;
    switch (csched_type) {
    case SYR_SCH_AC :
        SYR_cschedule_AC(len_table, &csched -> s_mark, &csched -> s_num_stages);
        transpose(len_table, SYR_nproc);
        SYR_cschedule_AC(len_table, &csched -> g_mark, &csched -> g_num_stages);
        break;

    case SYR_SCH_LP :
        SYR_cschedule_LP(len_table, &csched -> s_mark, &csched -> s_num_stages);
        transpose(len_table, SYR_nproc);
        SYR_cschedule_LP(len_table, &csched -> g_mark, &csched -> g_num_stages);
        break;

    case SYR_SCH_RS_N :
        SYR_cschedule_RS_N(len_table, &csched -> s_mark, &csched -> s_num_stages);
        transpose(len_table, SYR_nproc);
        SYR_cschedule_RS_N(len_table, &csched -> g_mark, &csched -> g_num_stages);
        break;

    case SYR_SCH_RS_NLP :
        /* This function is machine dependent */
        SYR_cschedule_RS_NLP(len_table, &csched -> s_mark, &csched -> s_num_stages);
        transpose(len_table, SYR_nproc);
        SYR_cschedule_RS_NLP(len_table, &csched -> g_mark, &csched -> g_num_stages);
        break;

    case SYR_SCH_NLP :
        /* This function is machine dependent */
        SYR_cschedule_NLP(len_table, &csched -> s_mark, &csched -> s_num_stages);
        transpose(len_table, SYR_nproc);
        SYR_cschedule_NLP(len_table, &csched -> g_mark, &csched -> g_num_stages);
        break;

    default :
        csched_err = TRUE;
        break;
    }

    if (csched_err) {
        free(csched);
        csched = NULL;
    }
    else {
        /* Set up send_n, fetch_n, tot_send_n, tot_fetch_n from len_tab */
        aloc_ary(csched -> send_n, int, SYR_nproc);
        aloc_ary(csched -> fetch_n, int, SYR_nproc);
        csched -> n_msgs_to_recv = csched -> n_msgs_to_send = 0;
        csched -> tot_send_n = csched -> tot_fetch_n = 0;
        csched -> buffer_size = 0;
        len_table[SYR_myid][SYR_myid] = self_len;
        for (i = 0; i < SYR_nproc; i++) {
            if ((csched -> send_n[i] = len_table[SYR_myid][i]) > 0) {
                csched -> n_msgs_to_send++;
                csched -> tot_send_n += csched -> send_n[i];
            }
            if (csched -> send_n[i] > csched -> buffer_size) {
                csched -> buffer_size = csched -> send_n[i];
            }
            if ((csched -> fetch_n[i] = len_table[i][SYR_myid]) > 0) {
                csched -> n_msgs_to_recv++;
                csched -> tot_fetch_n += csched -> fetch_n[i];
            }
            if (csched -> fetch_n[i] > csched -> buffer_size) {
                csched -> buffer_size = csched -> fetch_n[i];
            }
        }
    }
    /* Free the memory of length table */
    free_mat(len_table);
    SYR_dbg_exit();
    return (csched);
}

/*
 * The following procedures schedule the order of sending/receiving
 * data for gathering/scattering data in different stratages.
 */

/*
 * Scheduling of Asynchronous Communication.
 */

static void SYR_cschedule_AC(len_tab, mark, num_stages)
int **len_tab;
SYR_comm_stage **mark;
int *num_stages;
{
    SYR_comm_stage *stage_p, *stage_q;
    int partner;
    int i;
    int p1, p2, *perm;

    SYR_dbg_enter("SYR_cschedule_AC");
    /* Disturb the order of AC sending */
    aloc_ary(perm, int, SYR_nproc);
    for (i = 0; i < SYR_nproc; i++) {
        perm[i] = i;
    }
    /* Disturb the order of send/recv */
    for (i = 0; i < disturb_t(SYR_nproc); i++) {
        p1 = (int) (SYR_rand() * SYR_nproc);
        p2 = (int) (SYR_rand() * SYR_nproc);
        if (p1 != p2) {
            partner = perm[p1];
            perm[p1] = perm[p2];
            perm[p2] = partner;
        }
    }
    stage_q = stage_p = NULL;
    *num_stages = 0;
    for (i = 0; i < SYR_nproc; i++) {
        partner = perm[i];
        if ((len_tab[SYR_myid][partner] > 0) ||
            (len_tab[partner][SYR_myid] > 0)) {
            aloc_ary(stage_p, SYR_comm_stage, 1);
            stage_p -> s_partner = (len_tab[SYR_myid][partner] > 0) ?
                partner : SYR_INVALID_NODE;
            stage_p -> s_len = len_tab[SYR_myid][partner];
            stage_p -> s_offset = 0;
            stage_p -> r_partner = (len_tab[partner][SYR_myid] > 0) ?
                partner : SYR_INVALID_NODE;
            stage_p -> r_len = len_tab[partner][SYR_myid];
            stage_p -> r_offset = 0;
            stage_p -> next = stage_q;
            stage_q = stage_p;
            (*num_stages)++;
        }
    }
    *mark = stage_p;
    free(perm);

    SYR_dbg_exit();
}

/*
 * Scheduling of Linear Permutation.
 */

static void SYR_cschedule_LP(len_tab, mark, num_stages)
int **len_tab;
SYR_comm_stage **mark;
int *num_stages;
{
    SYR_comm_stage *stage_p, *stage_q;
    int partner;
    int i;

    SYR_dbg_enter("SYR_cschedule_LP");
    stage_q = stage_p = NULL;
    for (i = SYR_nproc - 1; i >= 1; i--) {
        partner = SYR_myid ^ i;
        aloc_ary(stage_p, SYR_comm_stage, 1);
        stage_p -> s_partner = (len_tab[SYR_myid][partner] > 0) ?
            partner : SYR_INVALID_NODE;
        stage_p -> s_len = len_tab[SYR_myid][partner];
        stage_p -> s_offset = 0;
        stage_p -> r_partner = (len_tab[partner][SYR_myid] > 0) ?
            partner : SYR_INVALID_NODE;
        stage_p -> r_len = len_tab[partner][SYR_myid];
        stage_p -> r_offset = 0;
        stage_p -> next = stage_q;
        stage_q = stage_p;
    }
    *mark = stage_p;
    *num_stages = SYR_nproc - 1;
    SYR_dbg_exit();
}

/*
 * Scheduling of Random Scheduling avoiding Node contention.
 */


static void SYR_cschedule_RS_N(len_tab, mark, num_stages)
int **len_tab;
SYR_comm_stage **mark;
int *num_stages;
{
    int i, j, dest, src;
    boolean done;
    int *ts_mask, *tr_mask, *ts_len, *tr_len;
    int *ccmap_ptr;
    int **ccmap;
    SYR_comm_stage *cp;

    SYR_dbg_enter("SYR_cschedule_RS_N");

    /* Allocate and build ccmap table */
    aloc_mat(ccmap, int , SYR_nproc, SYR_nproc);
    aloc_ary(ccmap_ptr, int, SYR_nproc);
    SYR_build_ccmap(len_tab, ccmap, ccmap_ptr);
    /* Allocate temp variables */
    aloc_ary(ts_mask, int, SYR_nproc);
    aloc_ary(tr_mask, int, SYR_nproc);
    aloc_ary(ts_len, int, SYR_nproc);
    aloc_ary(tr_len, int, SYR_nproc);

    *mark = NULL;
    *num_stages = 0;
    do {
        /* initialize mask arrays */
        for (i = 0; i < SYR_nproc; i++) {
            ts_mask[i] = tr_mask[i] = SYR_INVALID_NODE;
            ts_len[i] = tr_len[i] = 0;
        }

        done = TRUE;
        /* randomly choose starting node */
        src = (int) (SYR_rand() * SYR_nproc); 
        for (i = 0; i < SYR_nproc; i++) {
            j = 0;
            while ((j <= ccmap_ptr[src]) && (ts_mask[src] < 0)) {
                dest = ccmap[src][j];
                if ((dest >= 0) && (tr_mask[dest] < 0)) {
                    ts_mask[src] = dest;
                    tr_mask[dest] = src;
                    tr_len[dest] = ts_len[src] = len_tab[src][dest];
                    done = FALSE;   /* not yet, more work to do */
                    ccmap[src][j] = ccmap[src][ccmap_ptr[src]];
                    ccmap_ptr[src]--;
                }
                else {
                    j++;
                }
            }
            src = (src + 1) % SYR_nproc;
        }

        if (done == FALSE) {
            /* Some one has communication in this stage, add this */
            /* stage to communication stage list                  */
            aloc_ary(cp, SYR_comm_stage, 1);
            cp -> s_partner = ts_mask[SYR_myid];
            cp -> s_len = ts_len[SYR_myid];
            cp -> s_offset = 0;
            cp -> r_partner = tr_mask[SYR_myid];
            cp -> r_len = tr_len[SYR_myid];
            cp -> r_offset = 0;
            cp -> next = *mark;
            *mark = cp;
            (*num_stages)++;
        }
    } while (done == FALSE);

    /* We generate the communication stage list in reverse order, */
    /* reverse it again to get the origional order.               */
    *mark = SYR_reverse_comm_stg(*mark);
    free(tr_len);
    free(ts_len);
    free(tr_mask);
    free(ts_mask);
    free(ccmap_ptr);
    free_mat(ccmap);
    SYR_dbg_exit();
}

/*
 * This procedure free up the memory of unused scheduling data 
 * structure of communication.
 */

void SYR_comm_free_csched(csched)
SYR_com_pattern *csched;
{
    int i;
    SYR_comm_stage *stage_p, *stage_q;

    SYR_dbg_enter("SYR_comm_free_csched");
    if (csched != NULL) {
        free(csched -> send_list[0]);
        free(csched -> send_list);
        free(csched -> send_n);
        free(csched -> fetch_list[0]);
        free(csched -> fetch_list);
        free(csched -> fetch_n);
        stage_p = csched -> g_mark;
        while (stage_p != NULL) {
            stage_q = stage_p;
            stage_p = stage_p -> next;
            free(stage_q);
        }
        stage_p = csched -> s_mark;
        while (stage_p != NULL) {
            stage_q = stage_p;
            stage_p = stage_p -> next;
            free(stage_q);
        }
        free(csched);
    }
    SYR_dbg_exit();
}

/*
 * compress COM into CCOM 
 */

void SYR_build_ccmap(cmap, ccmap, ccmap_ptr)
int **cmap, **ccmap, *ccmap_ptr;
{
    int i, j;
    int ent_ptr, itemp;
    int disturb_times, col1, col2;


    /* copy cmap to ccmap, only first entries columns are valueable */
    for (i = 0; i < SYR_nproc; i++) {
        ent_ptr = 0;
        for (j = 0; j < SYR_nproc; j++) {
            if (cmap[i][j] > 0) {
                ccmap[i][ent_ptr] = j;
                ent_ptr++;
            }
        }
        ccmap[i][ent_ptr] = -1;  /* delimiter, end of useful entries */
        ccmap_ptr[i] = ent_ptr - 1;

        /* random switch entries in ccmap to disturb the order */
        /* each try will have ccmap in different order */
        /* swap enough times - disturb_times */
        disturb_times = disturb_t(ent_ptr);
        for (j = 0; j < disturb_times; j++) {
            col1 = (int) (SYR_rand() * ent_ptr);
            col2 = (int) (SYR_rand() * ent_ptr);
            if (col1 != col2) {
                itemp = ccmap[i][col1];
                ccmap[i][col1] = ccmap[i][col2];
                ccmap[i][col2] = itemp;
            }
        }
    }
} 

/*
 * This function reverse the order of input communication
 * stage list and returns the pointer of tail as new head.
 */

SYR_comm_stage *SYR_reverse_comm_stg(head)
SYR_comm_stage *head;
{
    SYR_comm_stage *p, *q;

    SYR_dbg_enter("SYR_reverse_comm_stg");
    p = NULL;
    while (head != NULL) {
        q = head -> next;
        head -> next = p;
        p = head;
        head = q;
    }
    SYR_dbg_exit();
    return(p);
}

/*
 * This procedure change the base of message types of synchronizing
 * and data messages to prevent that two adjecent communication
 * calls use the same message type.
 *
 * All communication part functions need to do real communications must
 * call this function first.
 */

static void SYR_mtype_flip()
{
    SYR_dbg_enter("SYR_mtype_flip");
    if (SYR_mtype_odd) {
        SYR_mtype_msg = SYR_PORT_MAX_M_TYPE;
        SYR_mtype_odd = FALSE;
    }
    else {
        SYR_mtype_msg = SYR_PORT_MAX_M_TYPE - 1;
        SYR_mtype_odd = TRUE;
    }
    SYR_dbg_exit();
}

/*
 * Send/receive data by using the scheduled communication stage lists.
 */

void SYR_comm(from, to, stages, n_stage, csched_type, unit_len)
char **to, **from;
SYR_comm_stage *stages;
int n_stage, unit_len;
SYR_csched_type csched_type;
{
    SYR_comm_stage *mp;
    int i, j;
    int s_partner, r_partner;
    int *s_msg_id, *r_msg_id;
    char sync_buf[1];

    SYR_dbg_enter("SYR_comm");
    SYR_mtype_flip();
    switch (csched_type) {
    case SYR_SCH_AC :
        /* Non-staglized communication */
        aloc_ary(s_msg_id, int, n_stage);
        aloc_ary(r_msg_id, int, n_stage);

        /* Post all buffers to receive messages */
        for (mp = stages, i = 0; mp != NULL; mp = mp -> next) {
            if ((r_partner = mp -> r_partner) >= 0) {
                r_msg_id[i++] = SYR_async_recv(SYR_mtype_msg, 
                    &to[r_partner][mp -> r_offset], 
                    mp -> r_len * unit_len, r_partner);
            }
        }

        /* Send out all messages */
        for (mp = stages, j = 0; mp != NULL; mp = mp -> next) {
            if ((s_partner = mp -> s_partner) >= 0) {
                s_msg_id[j++] = SYR_async_send(SYR_mtype_msg, 
                    &from[s_partner][mp -> s_offset],
                    mp -> s_len * unit_len, s_partner);
            }
        }

        /* Wait for complete of message transfer */
        while (j > 0) {
            SYR_async_msg_wait(s_msg_id[--j]);
        }
        free(s_msg_id);
        while (i > 0) {
            SYR_async_msg_wait(r_msg_id[--i]);
        }
        free(r_msg_id);
        break;

    case SYR_SCH_LP :
    case SYR_SCH_RS_N :
    case SYR_SCH_RS_NLP :
    case SYR_SCH_NLP :
        /* Staglized communication */
        for (mp = stages; mp != NULL; mp = mp -> next) {
            s_partner = mp -> s_partner;
            r_partner = mp -> r_partner;
            if (r_partner >= 0) {
                if (s_partner >= 0) {
                    /* Both send and receive */
                    if (s_partner == r_partner) {
                        /* Pairwise exchange */
                        SYR_swap(SYR_mtype_msg,
                            &from[s_partner][mp -> s_offset],
                            mp -> s_len * unit_len, s_partner, 
                            SYR_mtype_msg, 
                            &to[r_partner][mp -> r_offset],
                            mp -> r_len * unit_len);
                    }
                    else {
                        /* Non-pairwise send and receive */
                        SYR_sync_send_recv(SYR_mtype_msg,
                            &from[s_partner][mp -> s_offset],
                            mp -> s_len * unit_len, s_partner,
                            SYR_mtype_msg, 
                            &to[r_partner][mp -> r_offset],
                            mp -> r_len * unit_len, r_partner);
                    }
                }
                else {
                    /* Only receive */
                    SYR_sync_recv(SYR_mtype_msg, &to[r_partner][mp -> r_offset],
                        mp -> r_len * unit_len, r_partner);
                }
            }
            else {
                if (s_partner >= 0) {
                    /* Only send */
                    SYR_sync_send(SYR_mtype_msg, &from[s_partner][mp -> s_offset],
                        mp -> s_len * unit_len, s_partner);
                }
                else {
                    /* Neither send nor receive. Do nothing */
                }
            }
        }
        break;

    default :
        SYR_panic("Error communication schedule type in communication.");
        break;
    }
    SYR_dbg_exit();
}


