/*-------------------------------------------------------------------
 * This file is written by Tseng-Hui Lin. If you have any question,
 * please send e-mail to:
 *     thlin@top.cis.syr.edu  or  thlin@nova.npac.syr.edu
 *
 * Last modify : 9/23/1993
 *-------------------------------------------------------------------
 */

/*
 * This file contains all definitions of variables and structures
 * used internally in communication part only.
 */

#ifndef COMM_INT_H
#define COMM_INT_H

#include <malloc.h>
#include "parti_comm.h"
#include "comm_md.h"

/* Constant */

#ifndef FALSE
    enum Boolean {FALSE, TRUE};
    typedef enum Boolean boolean;
#else FALSE
    typedef int boolean;
#endif FALSE

#ifndef NULL
#   define NULL (0)
#endif NULL

/* Allign communication buffer on double word (8 bytes) boundary */
#define SYR_MEM_ALIGN (8)
#define SYR_INVALID_NODE (-1)

/* Global data (Not changed after init, constant) */

extern int SYR_nproc, SYR_myid;

/* Global data (Variables) */

extern boolean SYR_mtype_odd;
extern int SYR_mtype_msg, SYR_mtype_sync;

/* Data structure */


/* Global functions */
/* File comm.c */
void SYR_comm(/* from, to, stages, n_stage, csched_type, unit_len */);
void SYR_build_ccmap(/* cmap, ccmap, ccmap_ptr */);
SYR_comm_stage *SYR_reverse_comm_stg(/* head */);

/* File commmd.c */
void SYR_machine_init(/*  */);
void SYR_cschedule_RS_NLP(/* len_tab, csched */);
void SYR_cschedule_NLP(/* len_tab, csched */);

/* tcomm.c */

/* File util.c */
void SYR_not_written(/*  */);
void SYR_panic(/* s */);
long SYR_srand(/* x */);
double SYR_rand(/*  */);

/* Global macros */

#ifdef DEBUG
#    define SYR_dbg_enter(s) trace_enter(s)
#    define SYR_dbg_exit() trace_exit()
#    define SYR_dbg_dump() trace_exit()
#else DEBUG
#    define SYR_dbg_enter(s)
#    define SYR_dbg_exit()
#    define SYR_dbg_dump()
#endif DEBUG

#define aloc_ary(p, type, size) \
{ \
    int _size; \
 \
    _size = ((size) <= 0) ? 1 : (size); \
    p = (type *) malloc(sizeof(type) * (_size)); \
    if ((p) == NULL) SYR_panic("memory allocation error "); \
}

#define disturb_t(nelem) (nelem * 3)

#define aloc_ary_a(p, type, size) \
{ \
    int _size; \
 \
    _size = ((size) <= 0) ? 1 : (size); \
    p = (type *) memalign(SYR_MEM_ALIGN, sizeof(type) * (_size)); \
    if ((p) == NULL) SYR_panic("memory allocatition error "); \
}

#define aloc_mat(name, type, nrow, ncol) \
{ \
    type *xtemp; \
    int xi, xj; \
 \
    aloc_ary(name, type *, nrow); \
    aloc_ary(xtemp, type, nrow * ncol); \
    for (xi = 0, xj = 0; xi < nrow; xi++, xj += ncol) { \
        name[xi] = &xtemp[xj]; \
    } \
}

#define free_mat(name) \
{ \
    free(name[0]); \
    free(name); \
}

#define mapmat(ptr, body, esize, dim) \
{ \
    int xi, xaccu; \
 \
    for (xi = 0, xaccu = 0; xi < dim; xi++) { \
        ptr[xi] = &body[xaccu]; \
        xaccu += esize[xi]; \
    } \
}

#define pack(buffer, data, map, length) \
{ \
    int _i; \
 \
    for (_i = 0; _i < length; _i++) buffer[_i] = data[map[_i]]; \
}

#define unpack(buffer, data, map, length, oper) \
{ \
    int _i; \
 \
    for (_i = 0; _i < length; _i++) data[map[_i]] oper buffer[_i]; \
}

#define multpack(data, buffer, map, length, repetition, dialation, base_shift, data_offset) \
{ \
    int _i, _j, _t1, _t2, _t3; \
 \
    for (_i = 0, _t1 = 0; _i < length; _i++) { \
        for (_j = 0, _t2 = _t1, _t3 = map[_i] * dialation + data_offset; \
            _j < repetition; _j++, _t2++, _t3 += base_shift) { \
            buffer[_t2] = data[_t3]; \
        } \
        _t1 += repetition; \
    } \
}

#define multpack2(data, buffer, map, map2, length, repetition, dialation, base_shift, data_offset) \
{ \
    int _i, _j, _t1, _t2, _t3; \
 \
    for (_i = 0, _t1 = 0; _i < length; _i++) { \
        for (_j = 0, _t2 = _t1, _t3 = map2[map[_i]] * dialation + data_offset; \
            _j < repetition; _j++, _t2++, _t3 += base_shift) { \
            buffer[_t2] = data[_t3]; \
        } \
        _t1 += repetition; \
    } \
}

#define multunpack(data, buffer, map, length, repetition, dialation, base_shift, data_offset, oper) \
{ \
    int _i, _j, _t1, _t2, _t3; \
 \
    for (_i = 0, _t1 = data_offset; _i < length; _i++) { \
        for (_j = 0, _t2 = _t1, _t3 = map[_i] * dialation; \
            _j < repetition; _j++, _t2++, _t3 += base_shift) { \
            data[_t3] oper buffer[_t2]; \
        } \
        _t1 += repetition; \
    } \
}

#endif COMM_INT_H
