/*-------------------------------------------------------------------
 * This file is written by Tseng-Hui Lin. If you have any question,
 * please send e-mail to:
 *     thlin@top.cis.syr.edu  or  thlin@nova.npac.syr.edu
 *
 * Last modify : 4/14/1993
 *-------------------------------------------------------------------
 */

#ifndef PARTI_COMM_H
#define PARTI_COMM_H

#include "port.h"

/* PARTI use 2 message types internally in SYR_mtype_flip() */
#define SYR_MAX_MSG_TYPE (SYR_PORT_MAX_M_TYPE - 2)

enum SYR_Rtn_code {SYR_RTN_SUCCESS, SYR_RTN_ERROR};
typedef enum SYR_Rtn_code SYR_rtn_code;

enum SYR_Sched_type {
    SYR_SCH_AC,        /* Asynchronous Communication */
    SYR_SCH_LP,        /* Linear Permutation */
    SYR_SCH_RS_N,      /* Random Scheduling avoiding Node contention */
    SYR_SCH_RS_NLP,    /* Random Scheduling avoiding Node & Link contention */
    SYR_SCH_NLP        /* avoiding Node & Link contention */
};
typedef enum SYR_Sched_type SYR_csched_type;

/* Data structure for communication stages */
typedef struct SYR_Comm_stage SYR_comm_stage;
struct SYR_Comm_stage {
    int s_partner, s_len, s_offset ;    /* specify send_partner */
    int r_partner, r_len, r_offset ;    /* recieve partner      */
    SYR_comm_stage *next;               /* Next stage           */
};

typedef struct com_pattern SYR_com_pattern;
struct com_pattern {
    int n_msgs_to_recv, n_msgs_to_send;
    int tot_send_n, *send_n, **send_list;
    int tot_fetch_n, *fetch_n, **fetch_list;
    SYR_csched_type csched_type;
    int g_num_stages, s_num_stages;
    SYR_comm_stage *g_mark, *s_mark;

    int nprocs, myproc, pid;
    int buffer_size;

    int ncolor, *gcolor;                 /*  To be junked */
    int *scolor;
};

void SYR_comm_init(/*  */);
SYR_csched_type SYR_set_csched_type(/* type */);
SYR_com_pattern *schedule_gather(/* n, local_list, maxdim */);
void SYR_comm_free_csched(/* csched */);

/* Debugging routines */
void SYR_seq_begin();
void SYR_seq_end();
void SYR_dump_ary(/* ary, len */);
void SYR_dump_mat(/* mat, row, cols */);
void SYR_dump_comm_stage(/* mark */);
void SYR_dump_com_ptn(/* ptn */);

void cgather(/* to, from, sched */);
void igather(/* to, from, sched */);
void fgather(/* to, from, sched */);
void dgather(/* to, from, sched */);
void cmultigather(/* to, from, n_scheds, sched, base_shift, dialation, repetition */);
void imultigather(/* to, from, n_scheds, sched, base_shift, dialation, repetition */);
void fmultigather(/* to, from, n_scheds, sched, base_shift, dialation, repetition */);
void dmultigather(/* to, from, n_scheds, sched, base_shift, dialation, repetition */);

void cscatter(/* from, to, sched */);
void iscatter(/* from, to, sched */);
void fscatter(/* from, to, sched */);
void dscatter(/* from, to, sched */);
void cmultiscatter(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);
void imultiscatter(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);
void fmultiscatter(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);
void dmultiscatter(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);

void cscatter_add(/* from, to, sched */);
void iscatter_add(/* from, to, sched */);
void fscatter_add(/* from, to, sched */);
void dscatter_add(/* from, to, sched */);
void cmultiscatter_add(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);
void imultiscatter_add(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);
void fmultiscatter_add(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);
void dmultiscatter_add(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);

void cscatter_sub(/* from, to, sched */);
void iscatter_sub(/* from, to, sched */);
void fscatter_sub(/* from, to, sched */);
void dscatter_sub(/* from, to, sched */);
void cmultiscatter_sub(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);
void imultiscatter_sub(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);
void fmultiscatter_sub(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);
void dmultiscatter_sub(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);

void cscatter_mult(/* from, to, sched */);
void iscatter_mult(/* from, to, sched */);
void fscatter_mult(/* from, to, sched */);
void dscatter_mult(/* from, to, sched */);
void cmultiscatter_mult(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);
void imultiscatter_mult(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);
void fmultiscatter_mult(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);
void dmultiscatter_mult(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);

void cscatternc(/* from, to, sched */);
void iscatternc(/* from, to, sched */);
void fscatternc(/* from, to, sched */);
void dscatternc(/* from, to, sched */);
void cmultiscatternc(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);
void imultiscatternc(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);
void fmultiscatternc(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);
void dmultiscatternc(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);

void cscatter_addnc(/* from, to, sched */);
void iscatter_addnc(/* from, to, sched */);
void fscatter_addnc(/* from, to, sched */);
void dscatter_addnc(/* from, to, sched */);
void cmultiscatter_addnc(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);
void imultiscatter_addnc(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);
void fmultiscatter_addnc(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);
void dmultiscatter_addnc(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);

void cscatter_subnc(/* from, to, sched */);
void iscatter_subnc(/* from, to, sched */);
void fscatter_subnc(/* from, to, sched */);
void dscatter_subnc(/* from, to, sched */);
void cmultiscatter_subnc(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);
void imultiscatter_subnc(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);
void fmultiscatter_subnc(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);
void dmultiscatter_subnc(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);

void cscatter_multnc(/* from, to, sched */);
void iscatter_multnc(/* from, to, sched */);
void fscatter_multnc(/* from, to, sched */);
void dscatter_multnc(/* from, to, sched */);
void cmultiscatter_multnc(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);
void imultiscatter_multnc(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);
void fmultiscatter_multnc(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);
void dmultiscatter_multnc(/* from, to, n_scheds, sched, base_shift, dialation, repetition */);

#endif PARTI_COMM_H
