/*-------------------------------------------------------------------
 * This file is written by Tseng-Hui Lin. If you have any question,
 * please send e-mail to:
 *     thlin@top.cis.syr.edu  or  thlin@nova.npac.syr.edu
 *
 * Last modify : 9/26/1993
 *-------------------------------------------------------------------
 */

#include <stdio.h> /*@*/
#include "parti.h"
#include "comm_int.h"

void PREFIXgather(to, from, sched)
TYPE *from, *to;
SCHED *sched;
{
    PREFIXmultigather(to, from, 1, &sched, 1, 1, 1);
}

void PREFIXmultigather(to, from, n_scheds, sched,
    base_shift, dialation, repetition)
TYPE *from, *to;
int n_scheds, base_shift, dialation, repetition;
SCHED **sched;
{
    int i, sched_i, step;
    int i_buf_len, o_buf_len;
    TYPE *i_buf, *o_buf;
    TYPE **i_bufp, **o_bufp;
    SYR_com_pattern *ptn;

    SYR_dbg_enter("PREFIXmultigather");
    for (i = 0, i_buf_len = o_buf_len = 0; i < n_scheds; i++) {
        ptn = sched[i] -> sr;
        if (i_buf_len < ptn -> tot_fetch_n) {
            i_buf_len = ptn -> tot_fetch_n;
        }
        if (o_buf_len < ptn -> tot_send_n) {
            o_buf_len = ptn -> tot_send_n;
        }
    }
    aloc_ary_a(i_buf, TYPE, i_buf_len * repetition);
    aloc_ary_a(o_buf, TYPE, o_buf_len * repetition);
    aloc_ary(i_bufp, TYPE *, SYR_nproc);
    aloc_ary(o_bufp, TYPE *, SYR_nproc);
    step = 0;
    for (sched_i = 0; sched_i < n_scheds; sched_i++) {
        ptn = sched[sched_i] -> sr;
        for (i = 0, i_buf_len = o_buf_len = 0; i < SYR_nproc; i++) {
            i_bufp[i] = &i_buf[i_buf_len];
            o_bufp[i] = &o_buf[o_buf_len];
            i_buf_len += ptn -> fetch_n[i] * repetition;
            o_buf_len += ptn -> send_n[i] * repetition;
        }
        
        multpack(from, o_buf, ptn -> send_list[0], ptn -> tot_send_n,
            repetition, dialation, base_shift, step);
        SYR_comm(o_bufp, i_bufp, ptn -> g_mark, ptn -> g_num_stages,
            ptn -> csched_type, sizeof(TYPE) * repetition);
        for (i = 0; i < ptn -> send_n[SYR_myid]; i++) {
            i_bufp[SYR_myid][i] = o_bufp[SYR_myid][i];
        }
        multunpack(to, i_buf, sched[sched_i] -> perm, ptn -> tot_fetch_n, 
            repetition, dialation, base_shift, step, =);
        step += ptn -> tot_fetch_n * repetition;
    }
    free(o_bufp);
    free(i_bufp);
    free(o_buf);
    free(i_buf);
    SYR_dbg_exit();
}

void PREFIXscatter(from, to, sched)
TYPE *from, *to;
SCHED *sched;
{
    PREFIXmultiscatter(from, to, 1, &sched, 1, 1, 1);
}

void PREFIXmultiscatter(from, to, n_scheds, sched,
    base_shift, dialation, repetition)
TYPE *from, *to;
int n_scheds, base_shift, dialation, repetition;
SCHED **sched;
{
    int i, sched_i, step;
    int i_buf_len, o_buf_len;
    TYPE *i_buf, *o_buf;
    TYPE **i_bufp, **o_bufp;
    SYR_com_pattern *ptn;

    SYR_dbg_enter("PREFIXmultiscatter");
    for (i = 0, i_buf_len = o_buf_len = 0; i < n_scheds; i++) {
        ptn = sched[i] -> sr;
        if (i_buf_len < ptn -> tot_send_n) {
            i_buf_len = ptn -> tot_send_n;
        }
        if (o_buf_len < ptn -> tot_fetch_n) {
            o_buf_len = ptn -> tot_fetch_n;
        }
    }
    aloc_ary_a(i_buf, TYPE, i_buf_len * repetition);
    aloc_ary_a(o_buf, TYPE, o_buf_len * repetition);
    aloc_ary(i_bufp, TYPE *, SYR_nproc);
    aloc_ary(o_bufp, TYPE *, SYR_nproc);
    step = 0;
    for (sched_i = 0; sched_i < n_scheds; sched_i++) {
        ptn = sched[sched_i] -> sr;
        for (i = 0, i_buf_len = o_buf_len = 0; i < SYR_nproc; i++) {
            i_bufp[i] = &i_buf[i_buf_len];
            o_bufp[i] = &o_buf[o_buf_len];
            i_buf_len += ptn -> send_n[i] * repetition;
            o_buf_len += ptn -> fetch_n[i] * repetition;
        }
        multpack(from, o_buf, sched[sched_i] -> perm, ptn -> tot_fetch_n,
            repetition, dialation, base_shift, step);
        SYR_comm(o_bufp, i_bufp, ptn -> s_mark, ptn -> s_num_stages,
            ptn -> csched_type, sizeof(TYPE) * repetition);
        for (i = 0; i < ptn -> send_n[SYR_myid]; i++) {
            i_bufp[SYR_myid][i] = o_bufp[SYR_myid][i];
        }
        multunpack(to, i_buf, ptn -> send_list[0], ptn -> tot_send_n, 
            repetition, dialation, base_shift, step, =);
        step += ptn -> tot_fetch_n * repetition;
    }
    free(o_bufp);
    free(i_bufp);
    free(o_buf);
    free(i_buf);
    SYR_dbg_exit();
}

void PREFIXscatter_add(from, to, sched)
TYPE *from, *to;
SCHED *sched;
{
    PREFIXmultiscatter_add(from, to, 1, &sched, 1, 1, 1);
}

void PREFIXmultiscatter_add(from, to, n_scheds, sched,
    base_shift, dialation, repetition)
TYPE *from, *to;
int n_scheds, base_shift, dialation, repetition;
SCHED **sched;
{
    int i, sched_i, step;
    int i_buf_len, o_buf_len;
    TYPE *i_buf, *o_buf;
    TYPE **i_bufp, **o_bufp;
    SYR_com_pattern *ptn;

    SYR_dbg_enter("PREFIXmultiscatter_add");
    for (i = 0, i_buf_len = o_buf_len = 0; i < n_scheds; i++) {
        ptn = sched[i] -> sr;
        if (i_buf_len < ptn -> tot_send_n) {
            i_buf_len = ptn -> tot_send_n;
        }
        if (o_buf_len < ptn -> tot_fetch_n) {
            o_buf_len = ptn -> tot_fetch_n;
        }
    }
    aloc_ary_a(i_buf, TYPE, i_buf_len * repetition);
    aloc_ary_a(o_buf, TYPE, o_buf_len * repetition);
    aloc_ary(i_bufp, TYPE *, SYR_nproc);
    aloc_ary(o_bufp, TYPE *, SYR_nproc);
    step = 0;
    for (sched_i = 0; sched_i < n_scheds; sched_i++) {
        ptn = sched[sched_i] -> sr;
        for (i = 0, i_buf_len = o_buf_len = 0; i < SYR_nproc; i++) {
            i_bufp[i] = &i_buf[i_buf_len];
            o_bufp[i] = &o_buf[o_buf_len];
            i_buf_len += ptn -> send_n[i] * repetition;
            o_buf_len += ptn -> fetch_n[i] * repetition;
        }
        multpack(from, o_buf, sched[sched_i] -> perm, ptn -> tot_fetch_n,
            repetition, dialation, base_shift, step);
        SYR_comm(o_bufp, i_bufp, ptn -> s_mark, ptn -> s_num_stages,
            ptn -> csched_type, sizeof(TYPE) * repetition);
        for (i = 0; i < ptn -> send_n[SYR_myid]; i++) {
            i_bufp[SYR_myid][i] = o_bufp[SYR_myid][i];
        }
        multunpack(to, i_buf, ptn -> send_list[0], ptn -> tot_send_n, 
            repetition, dialation, base_shift, step, +=);
        step += ptn -> tot_fetch_n * repetition;
    }
    free(o_bufp);
    free(i_bufp);
    free(o_buf);
    free(i_buf);
    SYR_dbg_exit();
}

void PREFIXscatter_sub(from, to, sched)
TYPE *from, *to;
SCHED *sched;
{
    PREFIXmultiscatter_sub(from, to, 1, &sched, 1, 1, 1);
}

void PREFIXmultiscatter_sub(from, to, n_scheds, sched,
    base_shift, dialation, repetition)
TYPE *from, *to;
int n_scheds, base_shift, dialation, repetition;
SCHED **sched;
{
    int i, sched_i, step;
    int i_buf_len, o_buf_len;
    TYPE *i_buf, *o_buf;
    TYPE **i_bufp, **o_bufp;
    SYR_com_pattern *ptn;

    SYR_dbg_enter("PREFIXmultiscatter_sub");
    for (i = 0, i_buf_len = o_buf_len = 0; i < n_scheds; i++) {
        ptn = sched[i] -> sr;
        if (i_buf_len < ptn -> tot_send_n) {
            i_buf_len = ptn -> tot_send_n;
        }
        if (o_buf_len < ptn -> tot_fetch_n) {
            o_buf_len = ptn -> tot_fetch_n;
        }
    }
    aloc_ary_a(i_buf, TYPE, i_buf_len * repetition);
    aloc_ary_a(o_buf, TYPE, o_buf_len * repetition);
    aloc_ary(i_bufp, TYPE *, SYR_nproc);
    aloc_ary(o_bufp, TYPE *, SYR_nproc);
    step = 0;
    for (sched_i = 0; sched_i < n_scheds; sched_i++) {
        ptn = sched[sched_i] -> sr;
        for (i = 0, i_buf_len = o_buf_len = 0; i < SYR_nproc; i++) {
            i_bufp[i] = &i_buf[i_buf_len];
            o_bufp[i] = &o_buf[o_buf_len];
            i_buf_len += ptn -> send_n[i] * repetition;
            o_buf_len += ptn -> fetch_n[i] * repetition;
        }
        multpack(from, o_buf, sched[sched_i] -> perm, ptn -> tot_fetch_n,
            repetition, dialation, base_shift, step);
        SYR_comm(o_bufp, i_bufp, ptn -> s_mark, ptn -> s_num_stages,
            ptn -> csched_type, sizeof(TYPE) * repetition);
        for (i = 0; i < ptn -> send_n[SYR_myid]; i++) {
            i_bufp[SYR_myid][i] = o_bufp[SYR_myid][i];
        }
        multunpack(to, i_buf, ptn -> send_list[0], ptn -> tot_send_n, 
            repetition, dialation, base_shift, step, -=);
        step += ptn -> tot_fetch_n * repetition;
    }
    free(o_bufp);
    free(i_bufp);
    free(o_buf);
    free(i_buf);
    SYR_dbg_exit();
}

void PREFIXscatter_mult(from, to, sched)
TYPE *from, *to;
SCHED *sched;
{
    PREFIXmultiscatter_mult(from, to, 1, &sched, 1, 1, 1);
}

void PREFIXmultiscatter_mult(from, to, n_scheds, sched,
    base_shift, dialation, repetition)
TYPE *from, *to;
int n_scheds, base_shift, dialation, repetition;
SCHED **sched;
{
    int i, sched_i, step;
    int i_buf_len, o_buf_len;
    TYPE *i_buf, *o_buf;
    TYPE **i_bufp, **o_bufp;
    SYR_com_pattern *ptn;

    SYR_dbg_enter("PREFIXmultiscatter_mult");
    for (i = 0, i_buf_len = o_buf_len = 0; i < n_scheds; i++) {
        ptn = sched[i] -> sr;
        if (i_buf_len < ptn -> tot_send_n) {
            i_buf_len = ptn -> tot_send_n;
        }
        if (o_buf_len < ptn -> tot_fetch_n) {
            o_buf_len = ptn -> tot_fetch_n;
        }
    }
    aloc_ary_a(i_buf, TYPE, i_buf_len * repetition);
    aloc_ary_a(o_buf, TYPE, o_buf_len * repetition);
    aloc_ary(i_bufp, TYPE *, SYR_nproc);
    aloc_ary(o_bufp, TYPE *, SYR_nproc);
    step = 0;
    for (sched_i = 0; sched_i < n_scheds; sched_i++) {
        ptn = sched[sched_i] -> sr;
        for (i = 0, i_buf_len = o_buf_len = 0; i < SYR_nproc; i++) {
            i_bufp[i] = &i_buf[i_buf_len];
            o_bufp[i] = &o_buf[o_buf_len];
            i_buf_len += ptn -> send_n[i] * repetition;
            o_buf_len += ptn -> fetch_n[i] * repetition;
        }
        multpack(from, o_buf, sched[sched_i] -> perm, ptn -> tot_fetch_n,
            repetition, dialation, base_shift, step);
        SYR_comm(o_bufp, i_bufp, ptn -> s_mark, ptn -> s_num_stages,
            ptn -> csched_type, sizeof(TYPE) * repetition);
        for (i = 0; i < ptn -> send_n[SYR_myid]; i++) {
            i_bufp[SYR_myid][i] = o_bufp[SYR_myid][i];
        }
        multunpack(to, i_buf, ptn -> send_list[0], ptn -> tot_send_n, 
            repetition, dialation, base_shift, step, *=);
        step += ptn -> tot_fetch_n * repetition;
    }
    free(o_bufp);
    free(i_bufp);
    free(o_buf);
    free(i_buf);
    SYR_dbg_exit();
}

void PREFIXscatternc(from, to, sched)
TYPE *from, *to;
SCHED *sched;
{
    PREFIXmultiscatternc(from, to, 1, &sched, 1, 1, 1);
}

void PREFIXmultiscatternc(from, to, n_scheds, sched,
    base_shift, dialation, repetition)
TYPE *from, *to;
int n_scheds, base_shift, dialation, repetition;
SCHED **sched;
{
    int i, sched_i, step;
    int i_buf_len, o_buf_len;
    TYPE *i_buf, *o_buf;
    TYPE **i_bufp, **o_bufp;
    SYR_com_pattern *ptn;

    SYR_dbg_enter("PREFIXmultiscatternc");
    for (i = 0, i_buf_len = o_buf_len = 0; i < n_scheds; i++) {
        ptn = sched[i] -> sr;
        if (i_buf_len < ptn -> tot_send_n) {
            i_buf_len = ptn -> tot_send_n;
        }
        if (o_buf_len < ptn -> tot_fetch_n) {
            o_buf_len = ptn -> tot_fetch_n;
        }
    }
    aloc_ary_a(i_buf, TYPE, i_buf_len * repetition);
    aloc_ary_a(o_buf, TYPE, o_buf_len * repetition);
    aloc_ary(i_bufp, TYPE *, SYR_nproc);
    aloc_ary(o_bufp, TYPE *, SYR_nproc);
    step = 0;
    for (sched_i = 0; sched_i < n_scheds; sched_i++) {
        ptn = sched[sched_i] -> sr;
        for (i = 0, i_buf_len = o_buf_len = 0; i < SYR_nproc; i++) {
            i_bufp[i] = &i_buf[i_buf_len];
            o_bufp[i] = &o_buf[o_buf_len];
            i_buf_len += ptn -> send_n[i] * repetition;
            o_buf_len += ptn -> fetch_n[i] * repetition;
        }
        multpack2(from, o_buf, sched[sched_i] -> perm, 
            sched[sched_i] -> local_refs, ptn -> tot_fetch_n,
            repetition, dialation, base_shift, step);
        SYR_comm(o_bufp, i_bufp, ptn -> s_mark, ptn -> s_num_stages,
            ptn -> csched_type, sizeof(TYPE) * repetition);
        for (i = 0; i < ptn -> send_n[SYR_myid]; i++) {
            i_bufp[SYR_myid][i] = o_bufp[SYR_myid][i];
        }
        multunpack(to, i_buf, ptn -> send_list[0], ptn -> tot_send_n, 
            repetition, dialation, base_shift, step, =);
        step += ptn -> tot_fetch_n * repetition;
    }
    free(o_bufp);
    free(i_bufp);
    free(o_buf);
    free(i_buf);
    SYR_dbg_exit();
}

void PREFIXscatter_addnc(from, to, sched)
TYPE *from, *to;
SCHED *sched;
{
    PREFIXmultiscatter_addnc(from, to, 1, &sched, 1, 1, 1);
}

void PREFIXmultiscatter_addnc(from, to, n_scheds, sched,
    base_shift, dialation, repetition)
TYPE *from, *to;
int n_scheds, base_shift, dialation, repetition;
SCHED **sched;
{
    int i, sched_i, step;
    int i_buf_len, o_buf_len;
    TYPE *i_buf, *o_buf;
    TYPE **i_bufp, **o_bufp;
    SYR_com_pattern *ptn;

    SYR_dbg_enter("PREFIXmultiscatter_addnc");
    for (i = 0, i_buf_len = o_buf_len = 0; i < n_scheds; i++) {
        ptn = sched[i] -> sr;
        if (i_buf_len < ptn -> tot_send_n) {
            i_buf_len = ptn -> tot_send_n;
        }
        if (o_buf_len < ptn -> tot_fetch_n) {
            o_buf_len = ptn -> tot_fetch_n;
        }
    }
    aloc_ary_a(i_buf, TYPE, i_buf_len * repetition);
    aloc_ary_a(o_buf, TYPE, o_buf_len * repetition);
    aloc_ary(i_bufp, TYPE *, SYR_nproc);
    aloc_ary(o_bufp, TYPE *, SYR_nproc);
    step = 0;
    for (sched_i = 0; sched_i < n_scheds; sched_i++) {
        ptn = sched[sched_i] -> sr;
        for (i = 0, i_buf_len = o_buf_len = 0; i < SYR_nproc; i++) {
            i_bufp[i] = &i_buf[i_buf_len];
            o_bufp[i] = &o_buf[o_buf_len];
            i_buf_len += ptn -> send_n[i] * repetition;
            o_buf_len += ptn -> fetch_n[i] * repetition;
        }
        multpack2(from, o_buf, sched[sched_i] -> perm,
            sched[sched_i] -> local_refs, ptn -> tot_fetch_n,
            repetition, dialation, base_shift, step);
        SYR_comm(o_bufp, i_bufp, ptn -> s_mark, ptn -> s_num_stages,
            ptn -> csched_type, sizeof(TYPE) * repetition);
        for (i = 0; i < ptn -> send_n[SYR_myid]; i++) {
            i_bufp[SYR_myid][i] = o_bufp[SYR_myid][i];
        }
        multunpack(to, i_buf, ptn -> send_list[0], ptn -> tot_send_n, 
            repetition, dialation, base_shift, step, +=);
        step += ptn -> tot_fetch_n * repetition;
    }
    free(o_bufp);
    free(i_bufp);
    free(o_buf);
    free(i_buf);
    SYR_dbg_exit();
}

void PREFIXscatter_subnc(from, to, sched)
TYPE *from, *to;
SCHED *sched;
{
    PREFIXmultiscatter_subnc(from, to, 1, &sched, 1, 1, 1);
}

void PREFIXmultiscatter_subnc(from, to, n_scheds, sched,
    base_shift, dialation, repetition)
TYPE *from, *to;
int n_scheds, base_shift, dialation, repetition;
SCHED **sched;
{
    int i, sched_i, step;
    int i_buf_len, o_buf_len;
    TYPE *i_buf, *o_buf;
    TYPE **i_bufp, **o_bufp;
    SYR_com_pattern *ptn;

    SYR_dbg_enter("PREFIXmultiscatter_subnc");
    for (i = 0, i_buf_len = o_buf_len = 0; i < n_scheds; i++) {
        ptn = sched[i] -> sr;
        if (i_buf_len < ptn -> tot_send_n) {
            i_buf_len = ptn -> tot_send_n;
        }
        if (o_buf_len < ptn -> tot_fetch_n) {
            o_buf_len = ptn -> tot_fetch_n;
        }
    }
    aloc_ary_a(i_buf, TYPE, i_buf_len * repetition);
    aloc_ary_a(o_buf, TYPE, o_buf_len * repetition);
    aloc_ary(i_bufp, TYPE *, SYR_nproc);
    aloc_ary(o_bufp, TYPE *, SYR_nproc);
    step = 0;
    for (sched_i = 0; sched_i < n_scheds; sched_i++) {
        ptn = sched[sched_i] -> sr;
        for (i = 0, i_buf_len = o_buf_len = 0; i < SYR_nproc; i++) {
            i_bufp[i] = &i_buf[i_buf_len];
            o_bufp[i] = &o_buf[o_buf_len];
            i_buf_len += ptn -> send_n[i] * repetition;
            o_buf_len += ptn -> fetch_n[i] * repetition;
        }
        multpack2(from, o_buf, sched[sched_i] -> perm,
            sched[sched_i] -> local_refs, ptn -> tot_fetch_n,
            repetition, dialation, base_shift, step);
        SYR_comm(o_bufp, i_bufp, ptn -> s_mark, ptn -> s_num_stages,
            ptn -> csched_type, sizeof(TYPE) * repetition);
        for (i = 0; i < ptn -> send_n[SYR_myid]; i++) {
            i_bufp[SYR_myid][i] = o_bufp[SYR_myid][i];
        }
        multunpack(to, i_buf, ptn -> send_list[0], ptn -> tot_send_n, 
            repetition, dialation, base_shift, step, -=);
        step += ptn -> tot_fetch_n * repetition;
    }
    free(o_bufp);
    free(i_bufp);
    free(o_buf);
    free(i_buf);
    SYR_dbg_exit();
}

void PREFIXscatter_multnc(from, to, sched)
TYPE *from, *to;
SCHED *sched;
{
    PREFIXmultiscatter_multnc(from, to, 1, &sched, 1, 1, 1);
}

void PREFIXmultiscatter_multnc(from, to, n_scheds, sched,
    base_shift, dialation, repetition)
TYPE *from, *to;
int n_scheds, base_shift, dialation, repetition;
SCHED **sched;
{
    int i, sched_i, step;
    int i_buf_len, o_buf_len;
    TYPE *i_buf, *o_buf;
    TYPE **i_bufp, **o_bufp;
    SYR_com_pattern *ptn;

    SYR_dbg_enter("PREFIXmultiscatter_multnc");
    for (i = 0, i_buf_len = o_buf_len = 0; i < n_scheds; i++) {
        ptn = sched[i] -> sr;
        if (i_buf_len < ptn -> tot_send_n) {
            i_buf_len = ptn -> tot_send_n;
        }
        if (o_buf_len < ptn -> tot_fetch_n) {
            o_buf_len = ptn -> tot_fetch_n;
        }
    }
    aloc_ary_a(i_buf, TYPE, i_buf_len * repetition);
    aloc_ary_a(o_buf, TYPE, o_buf_len * repetition);
    aloc_ary(i_bufp, TYPE *, SYR_nproc);
    aloc_ary(o_bufp, TYPE *, SYR_nproc);
    step = 0;
    for (sched_i = 0; sched_i < n_scheds; sched_i++) {
        ptn = sched[sched_i] -> sr;
        for (i = 0, i_buf_len = o_buf_len = 0; i < SYR_nproc; i++) {
            i_bufp[i] = &i_buf[i_buf_len];
            o_bufp[i] = &o_buf[o_buf_len];
            i_buf_len += ptn -> send_n[i] * repetition;
            o_buf_len += ptn -> fetch_n[i] * repetition;
        }
        multpack2(from, o_buf, sched[sched_i] -> perm,
            sched[sched_i] -> local_refs, ptn -> tot_fetch_n,
            repetition, dialation, base_shift, step);
        SYR_comm(o_bufp, i_bufp, ptn -> s_mark, ptn -> s_num_stages,
            ptn -> csched_type, sizeof(TYPE) * repetition);
        for (i = 0; i < ptn -> send_n[SYR_myid]; i++) {
            i_bufp[SYR_myid][i] = o_bufp[SYR_myid][i];
        }
        multunpack(to, i_buf, ptn -> send_list[0], ptn -> tot_send_n, 
            repetition, dialation, base_shift, step, *=);
        step += ptn -> tot_fetch_n * repetition;
    }
    free(o_bufp);
    free(i_bufp);
    free(o_buf);
    free(i_buf);
    SYR_dbg_exit();
}
