/*-------------------------------------------------------------------
 * This file is written by Tseng-Hui Lin. If you have any question,
 * please send e-mail to:
 *     thlin@top.cis.syr.edu  or  thlin@nova.npac.syr.edu
 *
 * Last modify : 4/14/1993
 *-------------------------------------------------------------------
 */

/*
 * This file is machine dependent
 */

#include "../SYRCOMM/comm_int.h"
#include "comm_md.h"

static int SYR_ndim;
static int *SYR_link_order = NULL;

static boolean SYR_path_check(/* send, recv, stage_cnt, path_cube */);
static void SYR_path_mark(/* send, recv, stage_cnt, path_cube */);
static void SYR_mk_lk_order(/*  */);
static boolean SYR_mark_sr_NLP(/* len_tab, ts_mask, tr_mask,
    ts_len, tr_len, t_prev, t_next, path, path_p, path_cube */);

/*
 * This is the initial program of machine dependent part.
 */

void SYR_machine_init()
{
    SYR_dbg_enter("SYR_machine_init");
    SYR_ndim = SYR_nodedim();
    SYR_dbg_exit();
}

/*
 * fill out send & receive information for RS_NLP
 */

#define mark_sr(send, recv, loc) \
{ \
    ts_mask[send] = recv; \
    tr_mask[recv] = send; \
    tr_len[recv] = ts_len[send] = len_tab[send][recv]; \
    ccmap[send][loc] = ccmap[send][ccmap_ptr[send]]; \
    lcmap[send][ccmap[send][loc]] = loc; \
    lcmap[send][recv] = -1;  /* null this entry */ \
    ccmap_ptr[send]--; \
}

/*
 * for iPSC/2 and iPSC/860, use maximum pairwise exchange, no 
 * node/link contention 
 */

void SYR_cschedule_RS_NLP(len_tab, mark, num_stages)
int **len_tab;
SYR_comm_stage **mark;
int *num_stages;
{
    int i, j, dest;
    int src, stage_cnt;
    int *ts_mask, *tr_mask, *ts_len, *tr_len;
    int **ccmap, *ccmap_ptr, **lcmap, **path_cube;
    int loc, distance;
    boolean done, find_pair;
    SYR_comm_stage *cp;

    SYR_dbg_enter("SYR_cschedule_RS_NLP");

    /* Allocate and build ccmap table */
    aloc_mat(ccmap, int , SYR_nproc, SYR_nproc);
    aloc_ary(ccmap_ptr, int, SYR_nproc);
    SYR_build_ccmap(len_tab, ccmap, ccmap_ptr);
    /* Allocate temp variables */
    aloc_ary(ts_mask, int, SYR_nproc);
    aloc_ary(tr_mask, int, SYR_nproc);
    aloc_ary(ts_len, int, SYR_nproc);
    aloc_ary(tr_len, int, SYR_nproc);
    aloc_mat(lcmap, int , SYR_nproc, SYR_nproc);

    aloc_mat(path_cube, int , SYR_nproc, SYR_ndim);
    for (i = 0; i < SYR_nproc; i++) {
        /* keep track the location of every active entry */
        for (j = 0; j < SYR_nproc; j++) {
            lcmap[i][j] = -1;
        }
        for (j = 0; j <= ccmap_ptr[i]; j++) {
            lcmap[i][ccmap[i][j]] = j;
        }
        /* init path_cube[][] */
        for (j = 0; j < SYR_ndim; j++) {
            path_cube[i][j] = -1;
        }
    }

    *mark = NULL;
    *num_stages = 0;
    stage_cnt = 0;
    do {
        done = TRUE;

        /* initialize mask arrays */
        for (i = 0; i < SYR_nproc; i++) {
            ts_mask[i] = tr_mask[i] = SYR_INVALID_NODE;
            tr_len[i] = ts_len[i] = 0;
        }

        /* first round, find pairwise exchange */
        /* Note: No node contension here */
        src = (int) (SYR_rand() * SYR_nproc); /* random start */
        i = 0;
        find_pair = FALSE;
        while ((i < SYR_nproc) && (!find_pair)) {
            j = 0;
            while ((j <= ccmap_ptr[src]) && (!find_pair)) {
                dest = ccmap[src][j];
                if ((lcmap[src][dest] >= 0) && (lcmap[dest][src] >= 0)) {
                    find_pair = TRUE;
                }
                j++;
            }
            i++;
            src = (src + 1) % SYR_nproc;
        }
        if (find_pair) {
            /* Find a pair, src back to previous src */
            src = (src - 1 + SYR_nproc) % SYR_nproc;
            distance = src ^ dest;
            /* src --> dest */
            SYR_path_mark(src, dest, stage_cnt, path_cube);
            mark_sr(src, dest, lcmap[src][dest]);
            /* dest --> src */
            SYR_path_mark(dest, src, stage_cnt, path_cube);
            mark_sr(dest, src, lcmap[dest][src]);
            done = FALSE; /* not yet, more work to do */
        
            /* Find rest pairs with the same distance mask */
            src = (src + 1) % SYR_nproc;
            while (i < SYR_nproc) {
                /* use known and fixed distance in one permutation */
                dest = src ^ distance;
                if ((lcmap[src][dest] != -1) && (lcmap[dest][src] != -1)) {
                    /* src --> dest */
                    SYR_path_mark(src, dest, stage_cnt, path_cube);
                    mark_sr(src, dest, lcmap[src][dest]);
                    /* dest --> src */
                    SYR_path_mark(dest, src, stage_cnt, path_cube);
                    mark_sr(dest, src, lcmap[dest][src]);
                }  
                i++;
                src = (src + 1) % SYR_nproc;
            }
        }

        /* second round, fill out holes (Non-pair) */
        for (i = 0; i < SYR_nproc; i++) {
            j = 0;
            while((j <= ccmap_ptr[src]) && (ts_mask[src] < 0)) {
                dest = ccmap[src][j];
                if ((tr_mask[dest] < 0) &&
                    (SYR_path_check(src, dest, stage_cnt, path_cube))) {
                    SYR_path_mark(src, dest, stage_cnt, path_cube);
                    /* src --> dest */
                    mark_sr(src, dest, j);
                    if ((lcmap[dest][src] >= 0) &&
                        (tr_mask[src] < 0) && (ts_mask[dest] < 0) &&
                        (SYR_path_check(dest, src, stage_cnt, path_cube))) {
                        SYR_path_mark(dest, src, stage_cnt, path_cube);
                        /* dest --> src, got a pairwise exchange */
                        mark_sr(dest, src, lcmap[dest][src]);

                    }
                    done = FALSE;   /* not yet, more work to do */
                }
                else {
                    j++;
                }
            }
            src = (src + 1) % SYR_nproc;
        }

        if (done == FALSE) {
            /* Some one has communication in this stage, add this */
            /* stage to communication stage list                  */
            aloc_ary(cp, SYR_comm_stage, 1);
            cp -> s_partner = ts_mask[SYR_myid];
            cp -> s_len = ts_len[SYR_myid];
            cp -> s_offset = 0;
            cp -> r_partner = tr_mask[SYR_myid];
            cp -> r_len = tr_len[SYR_myid];
            cp -> r_offset = 0;
            cp -> next = *mark;
            *mark = cp;
            (*num_stages)++;
        }
        stage_cnt++;
    } while (done == FALSE) ;

    /* We generate the communication stage list in reverse order, */
    /* reverse it again to get the origional order.               */
    *mark = SYR_reverse_comm_stg(*mark);

    free_mat(path_cube);
    free_mat(lcmap);
    free(tr_len);
    free(ts_len);
    free(tr_mask);
    free(ts_mask);
    free(ccmap_ptr);
    free_mat(ccmap);
    SYR_dbg_exit();
}

/*
 * check a desired path on iPSC, return TRUE if no one uses it, FALSE
 * otherwise.
 */

static boolean SYR_path_check(send, recv, stage_cnt, path_cube)
int send, recv, stage_cnt, **path_cube;
{
    int i;
    int cur_node, path, mask;
    boolean no_conflict;

    cur_node = send;
    path = send ^ recv;
    for (i = 0, no_conflict = TRUE, mask = 1; (i < SYR_ndim) && (no_conflict);
        i++, mask <<= 1) {   /* calculate path */
        if ((path & mask) != 0) {
            if (path_cube[cur_node][i] == stage_cnt) {
                no_conflict = FALSE;
            }
            cur_node = cur_node ^ mask;
        }
    }
    return (no_conflict);
}

/* 
 * mark wanted path so no other communication can overlap this one 
 */

static void SYR_path_mark(send, recv, stage_cnt, path_cube)
int send, recv, stage_cnt, **path_cube;
{
    int i;
    int cur_node, path, mask;

    cur_node = send;
    path = send ^ recv;
    for (i = 0, mask = 1; i < SYR_ndim; i++, mask <<= 1) {   /* calculate path */
        if ((path & mask) != 0) {
            path_cube[cur_node][i] = stage_cnt;
            cur_node = cur_node ^ mask;
        }
    }
}

/*
 * Generate a table contains 0 to #proc - 1 and sorted by the # of 1's
 * in their binary form.
 */

static void SYR_mk_lk_order()
{
    int i, j;
    int mask, nbit;
    int *link_nbit, *link_ord_ptr;

    SYR_dbg_enter("SYR_mk_lk_order");
    aloc_ary(link_nbit, int, SYR_nproc);
    for (i = 0; i < SYR_nproc; i++) {
        /* count how many 1's in number */
        for (j = 0, mask = 1, nbit = 0; j < SYR_ndim; j++, mask <<= 1) {
            if (i & mask) {
                nbit++;
            }
        }
        link_nbit[i] = nbit;
    }

    aloc_ary(SYR_link_order, int, SYR_nproc);
    aloc_ary(link_ord_ptr, int, SYR_ndim + 1);
    /* Using PASCAL Trianglule to find the number of entry with i 1's. */
    for (i = 0; i <= SYR_ndim; i++) {
        link_ord_ptr[i] = 1;
    }
    for (i = 1; i < SYR_ndim; i++) {
        for (j = i; j > 0; j--) {
            link_ord_ptr[j] += link_ord_ptr[j - 1];
        }
    }
    /* There are link_ord_ptr[i] entries with i 1's */

    for (i = SYR_ndim - 1; i > 0; i--) {
        link_ord_ptr[i] += link_ord_ptr[i + 1];
    }
    for (i = 0; i < SYR_ndim; i++) {
        link_ord_ptr[i] = link_ord_ptr[i + 1];
    }
    link_ord_ptr[SYR_ndim] = 0;

    for (i = SYR_nproc - 1; i >= 0; i--) {
        nbit = link_nbit[i];
        SYR_link_order[link_ord_ptr[nbit]] = i;
        link_ord_ptr[nbit]++;
    }
    /* After the previous step, each pointer increses and equal */
    /* to its successor. We need to move the pointers back.     */
     for (i = 0; i < SYR_ndim; i++) {
        link_ord_ptr[i] = link_ord_ptr[i + 1];
    }
    link_ord_ptr[SYR_ndim] = 0;
    free(link_ord_ptr);
    free(link_nbit);
    SYR_dbg_exit();
}

/*
 * fill out send & receive information for NLP
 */

static boolean SYR_mark_sr_NLP(len_tab, ts_mask, tr_mask, 
    ts_len, tr_len, t_prev, t_next, path, path_p, path_cube)
int **len_tab;
int *ts_mask, *tr_mask, *ts_len, *tr_len, *t_prev, *t_next;
int path, path_p;
int **path_cube;
{
    int src, dest;
    boolean some_comm;

    SYR_dbg_enter("SYR_mark_sr_NLP");
    some_comm = FALSE;
    /* Fill send mask */
    src = t_next[SYR_nproc];
    while (src != SYR_nproc) {
        dest = src ^ path;
        if ((ts_mask[dest] < 0) && (tr_mask[dest] < 0) &&
            ((len_tab[dest][src] > 0) || (len_tab[src][dest] > 0))) { 
            if ((len_tab[src][dest] > 0) &&
                (SYR_path_check(src, dest, path_p, path_cube))) { 
                ts_len[src] = tr_len[dest] = len_tab[src][dest];
                len_tab[src][dest] = 0;
                ts_mask[src] = dest;
                tr_mask[dest] = src;
                SYR_path_mark(src, dest, path_p, path_cube);
            }
            if ((len_tab[dest][src] > 0) &&
                (SYR_path_check(dest, src, path_p, path_cube))) { 
                ts_len[dest] = tr_len[src] = len_tab[dest][src];
                len_tab[dest][src] = 0;
                ts_mask[dest] = src;
                tr_mask[src] = dest;
                SYR_path_mark(dest, src, path_p, path_cube);
            }
            if ((ts_mask[src] >= 0) || (ts_mask[dest] >= 0)) {
                t_next[t_prev[dest]] = t_next[dest];
                t_prev[t_next[dest]] = t_prev[dest];
                t_next[t_prev[src]] = t_next[src];
                t_prev[t_next[src]] = t_prev[src];
                some_comm = TRUE;
            }
        }
        /* We can still use t_next[src] because    */
        /* the value in t_next[src] is not changed */
        src = t_next[src];
    }
    return(some_comm);
}

/*
 * Scheduling of avoiding Node and Link contention on iPSC.
 */

void SYR_cschedule_NLP(len_tab, mark, num_stages)
int **len_tab;
SYR_comm_stage **mark;
int *num_stages;
{
    int i, j;
    int path_p;
    int mask1, mask2, path_mask;
    int path1, path2;
    int *is_sub_path;
    int *ts_mask, *ts_len, *tr_mask, *tr_len, *t_prev, *t_next;
    int **path_cube;
    int **work_len_tab;
    boolean some_comm;
    SYR_comm_stage *cp;

    SYR_dbg_enter("SYR_cschedule_NLP");
    if ( SYR_link_order == NULL) {
        /* We do it when this function is first time called */
        SYR_mk_lk_order();  /* make table for NLP */
    }
    /* We don't want to modify content of length table in this */
    /* subroutine, so we make a copy of length table.          */
    aloc_mat(work_len_tab, int, SYR_nproc, SYR_nproc);
    for (i = 0; i < SYR_nproc; i++) {
        for (j = 0; j < SYR_nproc; j++) {
            work_len_tab[i][j] = len_tab[i][j];
        }
    }

    path_mask = ~ (0XFFFFFFFF << SYR_ndim);
    aloc_ary(is_sub_path, int, SYR_nproc);
    for (i = 0; i < SYR_nproc; i++) {
        is_sub_path[i] = -1;
    }

    aloc_ary(ts_mask, int, SYR_nproc);
    aloc_ary(tr_mask, int, SYR_nproc);
    aloc_ary(ts_len, int, SYR_nproc);
    aloc_ary(tr_len, int, SYR_nproc);
    aloc_ary(t_prev, int, SYR_nproc + 1);
    aloc_ary(t_next, int, SYR_nproc + 1);

    aloc_mat(path_cube, int , SYR_nproc, SYR_ndim);
    for (i = 0; i < SYR_nproc; i++) {
        /* init cube_path[][] */
        for (j = 0; j < SYR_ndim; j++) {
            path_cube[i][j] = -1;
        }
    }

    *mark = NULL;
    *num_stages = 0;
    for (path_p = 0; path_p < SYR_nproc - 1; path_p++) {

        some_comm = FALSE;
        for (i = 0; i < SYR_nproc; i++) {
            ts_mask[i] = tr_mask[i] = SYR_INVALID_NODE;
            ts_len[i] = tr_len[i] = 0;
            /* Make a double linked array. */
            t_prev[i] = i - 1;
            t_next[i] = i + 1;
        }
        /* Use an extra one, the SYR_nproc'th, as head. */
        /* Make the double linked array circulated */
        t_prev[0] = SYR_nproc;
        t_prev[SYR_nproc] = SYR_nproc - 1;
        t_next[SYR_nproc] = 0;

        /* Find sub paths of given link */
        path1 = SYR_link_order[path_p];
        path2 = ~ path1;
        for (i = 0, mask1 = mask2 = path_mask; i < SYR_ndim;
            i++, mask1 = (mask1 << 1) & path_mask, mask2 = (mask2 >> 1) & path_mask) {
            is_sub_path[path1 & mask1] = is_sub_path[path1 & mask2] =
                is_sub_path[path2 & mask1] = is_sub_path[path2 & mask2] = path_p;
        }

        /* First run, current path and its sub-paths.     */
        /* There is no link contention between sub-paths. */
        for (i = path_p; i < SYR_nproc - 1; i++) {
            path1 = SYR_link_order[i];
            if (is_sub_path[path1] == path_p) {
                if (SYR_mark_sr_NLP(work_len_tab, ts_mask, tr_mask, ts_len, tr_len,
                    t_prev, t_next, path1, path_p, path_cube)) {
                    some_comm = TRUE;
                }
            }
        }

        /* Second run, fill up all possible link */
        for (i = path_p; i < SYR_nproc - 1; i++) {
            path1 = SYR_link_order[i];
            if (is_sub_path[path1] != path_p) {
                if (SYR_mark_sr_NLP(work_len_tab, ts_mask, tr_mask, ts_len, tr_len,
                    t_prev, t_next, path1, path_p, path_cube)) {
                    some_comm = TRUE;
                }
            }
        }

        if (some_comm) {
            /* Some one has communication in this stage, add this */
            /* stage to communication stage list                  */
            aloc_ary(cp, SYR_comm_stage, 1);
            cp -> s_partner = ts_mask[SYR_myid];
            cp -> s_len = ts_len[SYR_myid];
            cp -> s_offset = 0;
            cp -> r_partner = tr_mask[SYR_myid];
            cp -> r_len = tr_len[SYR_myid];
            cp -> r_offset = 0;
            cp -> next = *mark;
            *mark = cp;
            (*num_stages)++;
        }
    }

    /* We generate the communication stage list in reverse order, */
    /* reverse it again to get the origional order.               */
    *mark = SYR_reverse_comm_stg(*mark);

    free_mat(path_cube);
    free(t_next);
    free(t_prev);
    free(tr_len);
    free(ts_len);
    free(tr_mask);
    free(ts_mask);
    free(is_sub_path);
    SYR_dbg_exit();
}

