/*-------------------------------------------------------------------
 * This file is written by Tseng-Hui Lin. If you have any question,
 * please send e-mail to:
 *     thlin@top.cis.syr.edu  or  thlin@nova.npac.syr.edu
 *
 * Last modify : 4/14/1993
 *-------------------------------------------------------------------
 */

#include <stdio.h>
#include "parti.h"

#define MAX_DATA (64)
#define MAX_NODE (32)
#define MAX_BUFF (MAX_DATA * 4)
#define ID_OFF (10000)
#define D1_OFF (100)
#define DIM1 (10)

int com_patt[MAX_NODE][MAX_NODE];
int data[DIM1][MAX_BUFF], result[DIM1][MAX_BUFF], ans[DIM1][MAX_BUFF];
int proc[MAX_NODE][MAX_BUFF], indx[MAX_NODE][MAX_BUFF];
int myid, nproc;
int hiid;
int csched_type, rep_times, mat_dim;
double drand48();

#ifndef PARTI_COMM_H
/* For origional parti */
#include <cube.h>
#define SYR_SCH_AC (0)
#define SYR_SCH_LP (1)
#define SYR_SCH_RS_N (2)
#define SYR_SCH_RS_NLP (3)
#define SYR_SCH_NLP (4)
#define SYR_comm_init()
#define SYR_mynode() mynode();
#define SYR_numnodes() numnodes();
#define SYR_set_csched_type(csched_type) {csched_type = 1;}
#define SYR_gsync() gsync()
#endif SYR_SCH_AC

main(argc, argv)
int argc;
char *argv[];
{
    int i, j;

    SYR_comm_init();
    myid = SYR_mynode();
    nproc = SYR_numnodes();
    if (argc >= 2) {
        switch (atoi(argv[1])) {
        case 0 :
            csched_type = SYR_SCH_AC;
            break;
        case 1 :
            csched_type = SYR_SCH_LP;
            break;
        case 2 :
            csched_type = SYR_SCH_RS_N;
            break;
        case 3 :
            csched_type = SYR_SCH_RS_NLP;
            break;
        case 5 :
            csched_type = SYR_SCH_NLP;
            break;
        default :
            csched_type = -1;
            break;
        }
        if (csched_type >= 0) {
            SYR_set_csched_type(csched_type);
            hiid = myid * ID_OFF;
            for (i = 0; i < DIM1; i++) {
                for (j = 0; j < MAX_BUFF; j++) {
                    data[i][j] = hiid + i * D1_OFF + j;
                }
            }
            if (argc >= 3) {
                rep_times = atoi(argv[2]);
                if (rep_times <= 0) {
                    rep_times = 1;
                }
            }
            else {
                rep_times = 1;
            }
            for (i = 0; i < rep_times; i++) {
                do_test();
            }
        }
        else {
            fprintf(stderr, "Error comm type\n");
        }
    }
    else {
        fprintf(stderr, "Use : %s comm-sched-type repeat\n", argv[0]);
    }
}

do_test()
{
    int i, j, k;
    SCHED *sch;

    for (i = 0; i < MAX_NODE; i++) {
        for (j = 0; j < MAX_DATA; j++) {
            while ((proc[i][j] = (int) nproc * drand48()) == i) ;
            indx[i][j] = (int) MAX_DATA * drand48();
        }
    }
    seq_begin();
        printf("%2d Proc:", myid); 
        for (i = 0; i < MAX_DATA; i++) {
            printf("%2d ", proc[myid][i]);
        }
        printf("\n   Indx:");
        for (i = 0; i < MAX_DATA; i++) {
            printf("%2d ", indx[myid][i]);
        }
        printf("\n");
    seq_end();

    sch = schedule(indx[myid], proc[myid], MAX_DATA, DIM1);

    clr_result();
    igather(result, data, sch);
    for (i = 0; i < MAX_DATA; i++) {
        ans[0][i] = proc[myid][i] * ID_OFF + indx[myid][i];
    }
    chk_result("igather", 1);
            
    clr_result();
    imultigather(result, data, 1, &sch, MAX_BUFF, 1, DIM1 - 1);
    for (i = 0; i < DIM1 - 1; i++) {
        for (j = 0; j < MAX_DATA; j++) {
            ans[i][j] = proc[myid][j] * ID_OFF + i * D1_OFF + indx[myid][j];
        }
    }
    chk_result("imultigather", DIM1 - 1);
            
    clr_result();
    iscatter_add(data, result, sch);
    for (i = 0; i < nproc; i++) {
        for (j = 0; j < MAX_DATA; j++) {
            if (proc[i][j] == myid) {
                ans[0][indx[i][j]] += i * ID_OFF + j;
            }
        }
    }
    chk_result("iscatter_add", 1);
            
    clr_result();
    imultiscatter_add(data, result, 1, &sch, MAX_BUFF, 1, DIM1 - 1);
    for (i = 0; i < nproc; i++) {
        for (j = 0; j < MAX_DATA; j++) {
            if (proc[i][j] == myid) {
                for (k = 0; k < DIM1 - 1; k++) {
                    ans[k][indx[i][j]] += k * D1_OFF + i * ID_OFF + j;
                }
            }
        }
    }
    chk_result("imultiscatter_add", DIM1 - 1);
}

clr_result()
{
    int i, j;

    for (i = 0; i < DIM1; i++) {
        for (j = 0; j < MAX_BUFF; j++) {
            ans[i][j] = result[i][j] = -1;
        }
    }
}

chk_result(test_name, dim)
char *test_name;
int dim;
{
    int i, j;
    int rtn;

    rtn = 1;
    for (i = 0, rtn = 1; ((i < DIM1) && (rtn == 1)); i++) {
        for (j = 0; ((j < MAX_BUFF) && (rtn == 1)); j++) {
            if (ans[i][j] != result[i][j]) {
                rtn = 0;
            }
        }
    }

    if (rtn == 0) {
        if (myid == 0) {
            printf("Error on %s test\n", test_name);
        }
        seq_begin();
            printf("%2d:results by %s\n", test_name);
            for (i = 0; i < dim; i++) {
                for (j = 0; j < MAX_DATA; j++) {
                    printf("%6d ", result[i][j]);
                }
                printf("\n");
            }
            printf("%2d:results should be:\n");
            for (i = 0; i < dim; i++) {
                for (j = 0; j < MAX_DATA; j++) {
                    printf("%6d ", ans[i][j]);
                }
                printf("\n");
            }
        seq_end();
    }
    else {
        if (myid == 0) {
            printf("Pass %s test\n", test_name);
        }
    }
}

seq_begin()
{
    int i;

    for (i = 0; i <= myid; i++) {
        SYR_gsync();
    }
}

seq_end()
{
    int i;

    for (i = myid; i < nproc; i++) {
        SYR_gsync();
    }
}

