#include <stdio.h>
#include "parti.h"
#include "parti_more.h"

void execute_gather(pattern,buf_list,data,size)
struct com_pattern *pattern;
char **buf_list,*data;
int size;
{
  static int call_count = 0;
  int i,j;
  int *l;
  char *buffer;
  int node;
   
#ifdef DEBUG
  trace_enter("execute_gather");
#endif

  /* send data */
  buffer = (char *) malloc(pattern->buffer_size*size);
  for(i=0;i<pattern->nprocs;i++){
    if(pattern->send_n[i]>0){
      l = pattern->send_list[i];
      for(j=0;j<pattern->send_n[i];j++){
        bytecopy(&buffer[j*size],&data[l[j]*size],size);
      }  
      csend(call_count+GATHER_EXECUTOR_MSG,buffer,
             size*pattern->send_n[i],i,pattern->pid);
    }  
  }
  free(buffer);   
    
  /* recv data */
  for(i=0;i<pattern->n_msgs_to_recv;i++){
    cprobe(call_count+GATHER_EXECUTOR_MSG);
    node = infonode();
    if(buf_list[node]==NULL) {
      fprintf(stderr,"%d : Error in execute gather: buffer list has\n",
              mynode());
      fprintf(stderr,"%d : NULL in needed buffer slot\n",mynode());
      exit();
    }
    crecv(call_count+GATHER_EXECUTOR_MSG,
          buf_list[node],pattern->fetch_n[node]*size);
  }
  call_count++;
#ifdef DEBUG
  trace_exit();
#endif
}

void execute_scatter(pattern,buf_list,data,size)
struct com_pattern *pattern;
char **buf_list,*data;
int size;
{
  static int call_count = 0;
  int node,i,j;
  int *l;
  char *buffer;

#ifdef DEBUG
  trace_enter("execute_scatter");
#endif

  /* send data */
  for(i=0;i<pattern->nprocs;i++){
    if(pattern->fetch_n[i]>0){
      csend(call_count+SCATTER_EXECUTOR_MSG,buf_list[i],
            size*pattern->fetch_n[i],i,pattern->pid);
    }
  }
    
  /* recv data */
  buffer = (char *) malloc(pattern->buffer_size*size);
  for(i=0;i<pattern->n_msgs_to_send;i++){
    cprobe(call_count+SCATTER_EXECUTOR_MSG);
    node = infonode();
    crecv(call_count+SCATTER_EXECUTOR_MSG,buffer,pattern->buffer_size*size);
    l = pattern->send_list[node];
    for(j=0;j<pattern->send_n[node];j++){
      bytecopy(&data[l[j]*size],&buffer[j*size],size);
    }  
  }
  free(buffer);
  call_count++;
#ifdef DEBUG 
  trace_exit(); 
#endif    
}

struct com_pattern *schedule_gather(n,local_list,maxdim)
int *n,**local_list;
int maxdim;
{
  int i,j;
 
  int *a,*wrk,n_msgs_to_send; /* used in 1st part */
  int *l,msg_size,node,*send_n,**send_list;  /* used in second part */
  struct com_pattern *out_struct;
  int n_msgs_to_recv;
  int size, f_size, s_size;
  static int sched_call = 0;

#ifdef DEBUG 
  trace_enter("schedule_gather"); 
#endif 
  /* find how many messages to be sent in exec phase */
  wrk = (int *) malloc(sizeof(int)*numnodes());
  a = (int *) malloc(sizeof(int)*numnodes());
  for(i=0;i<numnodes();i++){
    a[i] = (n[i]>0);
  }  
  gisum(a,numnodes(),wrk);
  n_msgs_to_send = a[mynode()];
  free(a);
  free(wrk);

  /* send messages to those who must send messages in exec phase */
  n_msgs_to_recv = 0;
  for(i=0;i<numnodes();i++){
    if(n[i]>0){
      csend(TRANS_DATA_MSG,local_list[i],sizeof(int)*n[i],i,mypid());
      n_msgs_to_recv++;
    }
  }

  /* recv messages sent above */
  send_n = (int *) malloc(sizeof(int)*numnodes());
  send_list = (int **) malloc(sizeof(int *)*numnodes());
  for(i=0;i<numnodes();i++) send_n[i] = 0;
  for(i=0;i<n_msgs_to_send;i++){
    cprobe(TRANS_DATA_MSG);
    msg_size = infocount();
    node = infonode();
    send_n[node] = msg_size/sizeof(int);
    send_list[node] = (int *) malloc(msg_size);
    crecv(TRANS_DATA_MSG,send_list[node],msg_size);
  }

 
  /* put pointers into structre to be passed out */
  out_struct = (struct com_pattern *) malloc(sizeof(struct com_pattern));
  out_struct->buffer_size = 0;
  f_size = s_size =  0;
  for(i=0;i<numnodes();i++){
    if(out_struct->buffer_size<send_n[i]){
      out_struct->buffer_size = send_n[i];
     }
	 s_size += send_n[i];
  }
 	size = (s_size < f_size) ? f_size: s_size; 

  if (sched_call == 0)  {
	wbuffer = (struct WBuffer *) malloc (sizeof (struct WBuffer));
	wbuffer->msg_id = (int *) malloc(sizeof(int) * numnodes());
	wbuffer->buffer = (char *) malloc(sizeof(double) * (size*maxdim));
	wbuffer->max_size = size*maxdim;
  }
  else if (wbuffer->max_size < (size*maxdim)) {
	free(wbuffer->buffer); 
	wbuffer->buffer = (char *)malloc((size*maxdim)*sizeof(double)); 
	if (NULL == wbuffer->buffer)
	  fprintf(stderr,"ERROR:not enough memory -parti working buffer\n"),exit(0);
	/* realloc is expensive ?
	if (NULL == (char *)realloc(wbuffer->buffer, size*sizeof(double))) 
	  */
	wbuffer->max_size = size*maxdim;
  }
  out_struct->n_msgs_to_recv = n_msgs_to_recv;
  out_struct->n_msgs_to_send = n_msgs_to_send;
  out_struct->send_n = send_n;
  out_struct->send_list = send_list;
  out_struct->fetch_n = n;
  out_struct->fetch_list = local_list;
  out_struct->nprocs = numnodes();
  out_struct->myproc = mynode();
  out_struct->pid = mypid();


 sched_call++;
#ifdef DEBUG
  trace_exit();
#endif
 
  return(out_struct);
}  
