#include "parti.h"
/*
  wrapper for the schedule primitive and it returns the index of this
  particular schedule call
*/

extern TTABLE *get_table_ptr();
extern int register_schedule();
extern int register_table();
extern SCHED *get_sched_ptr();

int
  schedule_(local,proc,ndata,maxdim)
int *local, *proc, *ndata, *maxdim;
{
  int index;
  int i;
  SCHED *schedptr;

  for(i=0;i<*ndata;i++)
     local[i]--; 
  schedptr = schedule(local,proc,*ndata, *maxdim);
  index    = register_schedule(schedptr);
  for(i=0;i<*ndata;i++)
     local[i]++;
  return(index);
}

/*
  wrapper for the build_translation_table primitive and it returns the
   index of this particular call to build the table
*/

int
  build_translation_table_(type,index_array,ndata)
int *type, *index_array, *ndata;
{
  int index,i;
  TTABLE *tabptr;

  for(i=0;i<*ndata;i++)
     index_array[i]--; 
  tabptr = build_translation_table(*type,index_array,*ndata);
  index  = register_table(tabptr);
  for(i=0;i<*ndata;i++)
     index_array[i]++;
  return(index);
}


/*
  wrapper around the dereference function
*/

void 
  dereference_(tabindex,index_array,local,proc,ndata)
int *tabindex, *ndata;
int *index_array, *local, *proc;
{
  int i;
  TTABLE *tabptr;

  for(i=0;i<*ndata;i++)
     index_array[i]--;
  tabptr = (TTABLE *)get_table_ptr(*tabindex);
  dereference(tabptr,index_array,local,proc,*ndata);
  for(i=0;i<*ndata;i++){
     index_array[i]++;
     local[i]++;
  }
}

/* 
   This wrapper has been written for the primitive localize.
*/

void 
  localize_(tabindex,schedindex,global_refs,local_refs,ndata,n_off_proc,
                                 my_size,maxdim)
int *tabindex,*schedindex;
int *ndata,*n_off_proc,*my_size, *maxdim;
int *global_refs,*local_refs;
{
  SCHED *schptr;
  int index,off_proc,i;
  TTABLE *tabptr;

        
  for(i=0;i<*ndata;i++)
     global_refs[i]--;
  tabptr = (TTABLE *)get_table_ptr(*tabindex);
  localize(tabptr,&schptr,global_refs,local_refs,*ndata,&off_proc,
                                                     *my_size, *maxdim);
  *schedindex = register_schedule(schptr);
  *n_off_proc = off_proc;
  for(i=0;i<*ndata;i++){
     local_refs[i]++;
     global_refs[i]++;
  }
}


void 
  mlocalize_(tindex,sindex1,sindex2,global_refs,local_refs,ndata,
n_off_proc1,n_off_proc2,my_size,n_scheds,sarray, maxdim)
int *tindex,*sindex1,*sindex2;
int *global_refs,*local_refs;
int *ndata,*n_off_proc1,*n_off_proc2,*my_size, *maxdim;
int *n_scheds,*sarray;
{
  TTABLE *tabptr;
  SCHED *schptr1,*schptr2;
  SCHED **scharray;
  int index,off_proc1,off_proc2,i;

        
  for(i=0;i<*ndata;i++)
     global_refs[i]--;
  tabptr = (TTABLE *)get_table_ptr(*tindex);
  if(*n_scheds > 0){
    scharray = (SCHED **)malloc(sizeof(SCHED *)**n_scheds);
    for(i=0;i<*n_scheds;i++)
       scharray[i] = (SCHED *)get_sched_ptr(sarray[i]);
  }
  mlocalize(tabptr,&schptr1,&schptr2,global_refs,local_refs,*ndata,
                  &off_proc1,&off_proc2,*my_size,*n_scheds,scharray, *maxdim);
  
  *sindex1 = register_schedule(schptr1);
  *sindex2 = register_schedule(schptr2);
  *n_off_proc1 = off_proc1;
  *n_off_proc2 = off_proc2;
  for(i=0;i<*ndata;i++){
     local_refs[i]++;
     global_refs[i]++;
  }
}



/* 
   This wrapper has been written for the primitive reglocalize.
*/

void 
  reglocalize_(dist,size,schedindex,global_refs,local_refs,ndata,n_off_proc,
                                 my_size,maxdim)
int *schedindex, *dist, *size;
int *ndata,*n_off_proc,*my_size, *maxdim;
int *global_refs,*local_refs;
{
  SCHED *schptr;
  int hops,found;
  int index,off_proc,i;

        
  for(i=0;i<*ndata;i++)
     global_refs[i]--;
  reglocalize(*dist,*size,&schptr,
                   global_refs,local_refs,*ndata,&off_proc,*my_size, *maxdim);
  *schedindex = register_schedule(schptr);
  *n_off_proc = off_proc;
  for(i=0;i<*ndata;i++){
     local_refs[i]++;
     global_refs[i]++;
  }
}

/* 
   This wrapper has been written for the primitive regmlocalize.
*/

void 
  regmlocalize_(dist,size,sindex1,sindex2,global_refs,local_refs,ndata,
n_off_proc1,n_off_proc2,my_size,n_scheds,sarray, maxdim)
int *sindex1,*sindex2,*dist,*size;
int *global_refs,*local_refs;
int *ndata,*n_off_proc1,*n_off_proc2,*my_size, *maxdim;
int *n_scheds,*sarray;
{
  SCHED *schptr1,*schptr2;
  SCHED **scharray;
  int index,off_proc1,off_proc2,i;

        
  for(i=0;i<*ndata;i++)
     global_refs[i]--;
  if(*n_scheds > 0){
    scharray = (SCHED **)malloc(sizeof(SCHED *)**n_scheds);
    for(i=0;i<*n_scheds;i++)
       scharray[i] = (SCHED *)get_sched_ptr(sarray[i]);
  }
  regmlocalize(*dist,*size,&schptr1,&schptr2,global_refs,local_refs,*ndata,
                &off_proc1,&off_proc2,*my_size,*n_scheds,scharray, *maxdim);
  *sindex1 = register_schedule(schptr1);
  *sindex2 = register_schedule(schptr2);
  *n_off_proc1 = off_proc1;
  *n_off_proc2 = off_proc2;
  for(i=0;i<*ndata;i++){
     local_refs[i]++;
     global_refs[i]++;
  }
}


void 
  free_table_(tabindex)
int *tabindex;
{
  free_table((TTABLE *)get_table_ptr(*tabindex));
}

void
  free_sched_(schedindex)
int *schedindex;
{
    free_sched((SCHED *)get_sched_ptr(*schedindex));
}
